"""
Sendmail Commands and Files
===========================

This module contains the following parsers:

SendmailMC - file ``/etc/mail/sendmail.mc``
-------------------------------------------
"""

from insights.specs import Specs
from insights import parser
from insights.core import TextFileOutput


@parser(Specs.sendmail_mc)
class SendmailMC(TextFileOutput):
    """
    Class for parsing ``/etc/mail/sendmail.mc`` file.

    Sample Input::

        define(`confDEF_USER_ID', ``8:12'')dnl
        dnl define(`confAUTO_REBUILD')dnl
        define(`confTO_CONNECT', `1m')dnl
        define(`confTRY_NULL_MX_LIST', `True')dnl
        define(`confDONT_PROBE_INTERFACES', `True')dnl
        define(`PROCMAIL_MAILER_PATH', `/usr/bin/procmail')dnl
        define(`ALIAS_FILE', `/etc/aliases')dnl
        define(`STATUS_FILE', `/var/log/mail/statistics')dnl
        define(`UUCP_MAILER_MAX', `2000000')dnl
        define(`confUSERDB_SPEC', `/etc/mail/userdb.db')dnl
        define(`confPRIVACY_FLAGS', `authwarnings,novrfy,noexpn,restrictqrun')dnl
        define(`confAUTH_OPTIONS', `A')dnl

    Examples:
        >>> from insights.core.filters import add_filter
        >>> from insights.specs import Specs
        >>> add_filter(Specs.sendmail_mc, 'confTRY_NULL_MX_LIST')
        >>> type(sendmail_mc_obj)
        <class 'insights.parsers.sendmail.SendmailMC'>
        >>> sendmail_mc_obj.last_scan("null_mx_list_line", "confTRY_NULL_MX_LIST")
    """
    pass



#!/usr/bin/env python

import json
import os
import re
import sys
import argparse
import logging

from datetime import datetime
from itertools import chain
from collections import OrderedDict

import insights

from insights import load_packages, parse_plugins
from insights.core import dr, filters
from insights.core.spec_factory import RegistryPoint
from insights.specs import Specs

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)
logger = logging.getLogger(__name__)


def load_default_plugins():
    dr.load_components("insights.specs.default")
    dr.load_components("insights.parsers")
    dr.load_components("insights.combiners")


def apply_filters(_format, _plugins, output=None):
    load_default_plugins()

    if _format not in ("yaml", "json"):
        logger.error("Unsupported format: {0}".format(_format))
        return 1

    if not _plugins:
        logger.error("Provide plugins to load.")
        return 1

    load_packages(parse_plugins(_plugins))

    if _format == "yaml":
        yaml_path = output
        if not yaml_path:
            logger.info(
                "Output filters to '{0}'".format(
                    os.path.join(os.path.dirname(insights.__file__), filters._filename)
                )
            )
            filters.dump()
        else:
            logger.info("Output filters to '{0}'".format(yaml_path))
            with open(yaml_path, 'w') as fp:
                filters.dump(fp)

    if _format == "json":
        json_path = output
        if not json_path:
            logger.error("Provide uploader.json location to load and output.")
            return 1

        if not os.path.exists(json_path):
            logger.error("Provided '{0}' path does not exist.".format(json_path))
            return 1

        with open(json_path) as fp:
            uploader_json = json.load(fp, object_pairs_hook=OrderedDict)

        specs = sorted(vars(Specs))
        _filters = {}
        for spec in specs:
            s = getattr(Specs, spec)
            if isinstance(s, RegistryPoint):
                f = filters.get_filters(s)
                if f:
                    _filters[spec] = sorted(f)

        for spec in chain.from_iterable(uploader_json[i] for i in ("commands", "files", "globs")):
            if spec["symbolic_name"] in _filters:
                spec["pattern"] = _filters[spec["symbolic_name"]]

        uploader_json["version"] = datetime.now().isoformat()

        pattern = re.compile(",")
        output = "\n".join(
            pattern.sub(",", l) for l in json.dumps(uploader_json, indent=4).splitlines()
        )

        with open(json_path, "w") as fp:
            fp.write(output)
    return 0


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-f", "--format", help="Filters format.", default="yaml")
    parser.add_argument("-o", "--output", help="Ouput file.", default="")
    parser.add_argument(
        "-p", "--plugins", help="Comma-separated list without spaces of plugins.", default=""
    )
    args = parser.parse_args()

    ret = apply_filters(args.format, args.plugins, args.output)
    sys.exit(ret)


if __name__ == "__main__":
    main()



from insights.client.auto_config import set_auto_configuration
from mock.mock import Mock, patch


@patch("insights.client.auto_config.InsightsConnection")
def test_sat_branch_info_called(connection):
    '''
    When is_satellite is True, means we're on sat. get_branch_info should be called.
    '''
    config = Mock(base_url=None, upload_url=None, legacy_upload=False)
    set_auto_configuration(config, 'test.com:443/redhat_access', 'some_cert', None, True, False)
    connection.return_value.get_branch_info.assert_called_once()


@patch("insights.client.auto_config.InsightsConnection")
def test_rhsm_branch_info_not_called(connection):
    '''
    When is_satellite is False, means we're on direct RHSM. get_branch_info should not be called.
    '''
    config = Mock(base_url=None, upload_url=None, legacy_upload=False)
    set_auto_configuration(config, 'cert-api.access.redhat.com', None, None, False, False)
    connection.return_value.get_branch_info.assert_not_called()



import pytest

from mock.mock import Mock

from insights.core.exceptions import SkipComponent
from insights.specs.datasources.kernel import current_version, default_version
from insights.parsers.uname import Uname
from insights.parsers import uname
from insights.specs import Specs
from insights.tests import context_wrap

UNAME = """
Linux vm37-130.gsslab.pek2.redhat.com 5.14.0-160.el9.x86_64 #1 SMP PREEMPT_DYNAMIC Thu Aug 25 20:41:37 EDT 2022 x86_64 x86_64 x86_64 GNU/Linux
"""

UNAME_ERROR_BLANK = ""


def test_current_kernel_version():
    uname = Uname(context_wrap(UNAME))

    broker = {
        Uname: uname
    }
    result = current_version(broker)
    assert result is not None
    assert result == '5.14.0-160.el9.x86_64'


def test_current_kernel_version_without_uname():
    with pytest.raises(uname.UnameError) as e_info:
        current_version({Uname: uname.Uname(context_wrap(UNAME_ERROR_BLANK))})
    assert 'Empty uname line' in str(e_info.value)


def test_default_version():
    grubby_command = Mock()
    grubby_command.content = ['/boot/vmlinuz-4.18.0-305.el8.x86_64']
    broker = {Specs.grubby_default_kernel: grubby_command}
    result = default_version(broker)
    assert result is not None
    assert result == '4.18.0-305.el8.x86_64'


def test_default_version_with_error():
    grubby_command = Mock()
    grubby_command.content = ['grep: /boot/grub2/grubenv: Permission denied']
    broker = {Specs.grubby_default_kernel: grubby_command}
    with pytest.raises(SkipComponent):
        default_version(broker)



import doctest
from insights.parsers import cifs
from insights.parsers.cifs import CIFSDebugData
from insights.tests import context_wrap


CIFS_DEBUG_DATA_CONTENT = r"""
DFS origin full path: \\ad.abc.com\DFSROOT\ame1
DFS leaf full path:   \\CVXDFAVB1\DFSRoot\ame1
"""

CIFSDebugData.last_scan('test_dfs_origin_test', 'DFS origin')


def test_cifs_debug_data():
    cifs_dd_obj = CIFSDebugData(context_wrap(CIFS_DEBUG_DATA_CONTENT))

    assert cifs_dd_obj.test_dfs_origin_test
    assert cifs_dd_obj.test_dfs_origin_test.get('raw_line') == r"DFS origin full path: \\ad.abc.com\DFSROOT\ame1"


def test_doc():
    env = {
            'cifs_dd_obj': CIFSDebugData(context_wrap(CIFS_DEBUG_DATA_CONTENT))
          }
    failures, _ = doctest.testmod(cifs, globs=env)
    assert failures == 0



import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers.gcp_network_interfaces import GCPNetworkInterfaces
from insights.tests import context_wrap

GCP_NIC_1 = """
[{
  "accessConfigs":[
     {
        "externalIp":"34.67.41.222",
        "type":"ONE_TO_ONE_NAT"
     }
  ],
  "dnsServers":[
     "169.254.169.254"
  ],
  "forwardedIps":[],
  "gateway":"10.128.0.1",
  "ip":"10.128.0.3",
  "ipAliases":[],
  "mac":"42:01:0a:80:00:03",
  "mtu":1460,
  "network":"projects/564421043971/networks/default",
  "subnetmask":"255.255.240.0",
  "targetInstanceIps":[]
}]
"""
GCP_NIC_2 = '[]'

GCP_NIC_CURL_1 = """
curl: (7) Failed to connect to 169.254.169.254 port 80: Connection timed out
""".strip()
GCP_NIC_CURL_2 = """
curl: (7) couldn't connect to host
""".strip()
GCP_NIC_CURL_3 = """
curl: (28) connect() timed out!
""".strip()


def test_azure_instance_place_ab_other():
    with pytest.raises(SkipComponent):
        GCPNetworkInterfaces(context_wrap(GCP_NIC_CURL_1))

    with pytest.raises(SkipComponent):
        GCPNetworkInterfaces(context_wrap(GCP_NIC_CURL_2))

    with pytest.raises(SkipComponent):
        GCPNetworkInterfaces(context_wrap(GCP_NIC_CURL_3))

    with pytest.raises(SkipComponent):
        GCPNetworkInterfaces(context_wrap(''))


def test_gcp_public_ips():
    gcp_nics = GCPNetworkInterfaces(context_wrap(GCP_NIC_1))
    assert gcp_nics.public_ips == ["34.67.41.222"]
    assert gcp_nics.raw[0]["dnsServers"] == ["169.254.169.254"]

    gcp_nics = GCPNetworkInterfaces(context_wrap(GCP_NIC_2))
    assert gcp_nics.public_ips == []



from insights.tests import context_wrap
from insights.parsers.sysconfig import StonithSysconfig

STONITH_SYSCONFIG = """
retry=3
retry-sleep=2
verbose=yes    # optional
""".strip()


def test_sysconfig_sshd():
    result = StonithSysconfig(context_wrap(STONITH_SYSCONFIG))
    assert result["retry"] == '3'
    assert result.get("retry-sleep") == '2'
    assert result.get("verbose") == 'yes'



"""
PostfixMaster - file ``/etc/postfix/master.cf``
===============================================

"""
from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.postfix_master)
class PostfixMaster(Parser, list):
    """
    Class to parses the content of postfix configuration files ``/etc/postfix/master.cf``

    Typical content looks like::

        # ==========================================================================
        # service type  private unpriv  chroot  wakeup  maxproc command + args
        #               (yes)   (yes)   (no)    (never) (100)
        # ==========================================================================
        smtp      inet  n       -       n       -       -       smtpd
        pickup    unix  n       -       n       60      1       pickup
        cleanup   unix  n       -       n       -       0       cleanup
        qmgr      unix  n       -       n       300     1       qmgr
        #qmgr     unix  n       -       n       300     1       oqmgr
        tlsmgr    unix  -       -       n       1000?   1       tlsmgr
        rewrite   unix  -       -       n       -       -       trivial-rewrite
        bounce    unix  -       -       n       -       0       bounce
        defer     unix  -       -       n       -       0       bounce
        trace     unix  -       -       n       -       0       bounce
        verify    unix  -       -       n       -       1       verify
        flush     unix  n       -       n       1000?   0       flush
        proxymap  unix  -       -       n       -       -       proxymap
        proxywrite unix -       -       n       -       1       proxymap
        smtp      unix  -       -       n       -       -       smtp
        relay     unix  -       -       n       -       -       smtp
                -o syslog_name=postfix/$service_name
        #       -o smtp_helo_timeout=5 -o smtp_connect_timeout=5
        showq     unix  n       -       n       -       -       showq
        error     unix  -       -       n       -       -       error
        retry     unix  -       -       n       -       -       error
        discard   unix  -       -       n       -       -       discard
        local     unix  -       n       n       -       -       local
        virtual   unix  -       n       n       -       -       virtual
        lmtp      unix  -       -       n       -       -       lmtp
        anvil     unix  -       -       n       -       1       anvil
        scache    unix  -       -       n       -       1       scache

    Examples:
        >>> type(postfix_master)
        <class 'insights.parsers.postfix_conf.PostfixMaster'>
        >>> len(postfix_master)
        24
        >>> postfix_master[-1] == {'service': 'scache', 'type': 'unix', 'private': '-', 'unpriv': '-', 'chroot': 'n', 'wakeup': '-', 'maxproc': '1', 'command': 'scache'}
        True

    Raises:
        insights.core.exceptions.SkipComponent: if the ``/etc/postfix/master.cf`` is empty.
    """

    def parse_content(self, content):
        service_info = None
        heading = ['service', 'type', 'private', 'unpriv', 'chroot', 'wakeup', 'maxproc', 'command']
        for line in content:
            if not line.strip() or line.startswith('#'):
                continue

            if ('=' in line or '$' in line) and service_info:
                service_info['args'] = service_info.get('args', [])
                service_info['args'].append(line.strip())

            elif len(line.split()) == 8:
                line_sp = line.split()
                service_info = dict(zip(heading, line_sp))
                self.append(service_info)

        if not self:
            raise SkipComponent("Empty result")



import doctest
import pytest

from insights.parsers import watchdog, SkipComponent
from insights.parsers.watchdog import WatchDogConf, WatchDogLog
from insights.tests import context_wrap

WATCHDOG_CONF = """
# The retry-timeout and repair limit are used to handle errors in a
# more robust manner. Errors must persist for longer than this to
# action a repair or reboot, and if repair-maximum attempts are
# made without the test passing a reboot is initiated anyway.

retry-timeout          = 60
repair-maximum         = 1
realtime               = yes
priority               = 1
""".strip()

WATCHDOG_LOG = """
DEBUG:root:0

INFO:root:Executing: /usr/bin/sg_persist -n -i -k -d /dev/mapper/mpathb

DEBUG:root:0   PR generation=0xc2, 4 registered reservation keys follow:
    0xfdfe0001
    0xfdfe0001
    0xfdfe0000
    0xfdfe0000
""".strip()


def test_watchdog_conf():
    watchdog_conf = WatchDogConf(context_wrap(WATCHDOG_CONF))
    assert len(watchdog_conf) == 4
    assert watchdog_conf.get('retry-timeout') == '60'
    assert watchdog_conf.get('repair-maximum') == '1'
    assert watchdog_conf.get('realtime') == 'yes'
    assert watchdog_conf.get('priority') == '1'


def test_watchdog_skip():
    with pytest.raises(SkipComponent):
        WatchDogConf(context_wrap(''))


def test_watchdog_log():
    WatchDogLog.last_scan('register_line', 'registered reservation keys')
    watch_log = WatchDogLog(context_wrap(WATCHDOG_LOG))
    assert watch_log.register_line is not None
    assert watch_log.register_line.get('raw_message') == 'DEBUG:root:0   PR generation=0xc2, 4 registered reservation keys follow:'


def test_watchdog_log_exception():
    watch_log = WatchDogLog(context_wrap(WATCHDOG_LOG))
    with pytest.raises(RuntimeError):
        list(watch_log.get_after('2013-1-1'))


def test_doc():
    env = {
        "watchdog_conf_obj": WatchDogConf(context_wrap(WATCHDOG_CONF))
    }
    failed_count, _ = doctest.testmod(watchdog, globs=env)
    assert failed_count == 0



import pytest
from insights.core.ls_parser import FilePermissions
from insights.parsers.ls import FileListing, LSlHFiles
from insights.tests import context_wrap
from insights.tests.parsers.test_ls_file_listing import (
    LS_FILE_PERMISSIONS_DOC,
    MULTIPLE_DIRECTORIES,
)

PERMISSIONS_TEST_EXCEPTION_VECTORS = [
    ('-rw------ 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('bash: ls: command not found', True),
    ('-rw------ 1 root root 762 Se', True),
    ('-rw------- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------. 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------@ 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------+ 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------* 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------asdf 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw------- 1 ro:t root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('-rw------- 1 root r:ot 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('-rwasdfas- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('-rwx/----- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('/usr/bin/ls: cannot access /boot/grub2/grub.cfg: No such file or directory', True),
    ('cannot access /boot/grub2/grub.cfg: No such file or directory', True),
    ('No such file or directory', True),
    ('adsf', True),
]

PERMISSIONS_TEST_VECTORS = [
    (
        '-rw------- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config',
        False,
        'rw-',
        '---',
        '---',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        True,
        True,
    ),
    (
        '-rw------- 1 root root 762 Sep 23 002 /a path/with/spaces everywhere',
        False,
        'rw-',
        '---',
        '---',
        'root',
        'root',
        '/a path/with/spaces everywhere',
        True,
        True,
        True,
        True,
    ),
    (
        '-rw------- 1 root root 762 Sep 23 002 no_slash_here',
        False,
        'rw-',
        '---',
        '---',
        'root',
        'root',
        'no_slash_here',
        True,
        True,
        True,
        True,
    ),
    (
        '-rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config',
        False,
        'rw-',
        '---',
        '---',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        True,
        True,
    ),
    (
        '-rw-rw-rw-. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config',
        False,
        'rw-',
        'rw-',
        'rw-',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        False,
        False,
    ),
    (
        '-rw-rw---- 1 root user 762 Sep 23 002 /etc/ssh/sshd_config',
        True,
        'rw-',
        'rw-',
        '---',
        'root',
        'user',
        '/etc/ssh/sshd_config',
        True,
        False,
        False,
        False,
    ),
    (
        '-rw------- 1 root user 762 Sep 23 002 /etc/ssh/sshd_config',
        False,
        'rw-',
        '---',
        '---',
        'root',
        'user',
        '/etc/ssh/sshd_config',
        True,
        False,
        True,
        True,
    ),
    (
        '-rw------- 1 user root 762 Sep 23 002 /etc/ssh/sshd_config',
        False,
        'rw-',
        '---',
        '---',
        'user',
        'root',
        '/etc/ssh/sshd_config',
        False,
        False,
        False,
        False,
    ),
    (
        '-rw-rw---- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config',
        True,
        'rw-',
        'rw-',
        '---',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        True,
        True,
    ),
    (
        '-rw-rw-r-- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config',
        True,
        'rw-',
        'rw-',
        'r--',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        False,
        True,
    ),
    (
        '-rw-rw--w- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config',
        True,
        'rw-',
        'rw-',
        '-w-',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        True,
        False,
    ),
    (
        '---------- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config',
        False,
        '---',
        '---',
        '---',
        'root',
        'root',
        '/etc/ssh/sshd_config',
        True,
        True,
        True,
        True,
    ),
]


def test_permissions():
    for vector in PERMISSIONS_TEST_VECTORS:
        (
            line,
            with_group,
            permissions_owner,
            permissions_group,
            permissions_other,
            owner,
            group,
            path,
            owned_by_root_user,
            owned_by_root_user_and_group,
            only_root_can_read,
            only_root_can_write,
        ) = vector
        p = FilePermissions(line)
        assert p.perms_owner == permissions_owner
        assert p.perms_group == permissions_group
        assert p.perms_other == permissions_other
        assert p.owner == owner
        assert p.group == group
        assert p.owned_by('root', also_check_group=False) == owned_by_root_user
        assert p.owned_by('root', also_check_group=True) == owned_by_root_user_and_group
        assert p.only_root_can_read(root_group_can_read=with_group) == only_root_can_read
        assert p.only_root_can_write(root_group_can_write=with_group) == only_root_can_write
        assert p.all_zero() == all(
            (p.perms_owner == '---', p.perms_group == '---', p.perms_other == '---')
        )
        assert p.owner_can_read() == ('r' in p.perms_owner)
        assert p.owner_can_write() == ('w' in p.perms_owner)
        assert p.owner_can_only_read() == ('r--' == p.perms_owner)
        assert p.group_can_read() == ('r' in p.perms_group)
        assert p.group_can_write() == ('w' in p.perms_group)
        assert p.group_can_only_read() == ('r--' == p.perms_group)
        assert p.others_can_read() == ('r' in p.perms_other)
        assert p.others_can_write() == ('w' in p.perms_other)
        assert p.others_can_only_read() == ('r--' == p.perms_other)


def test_permissions_invalid():
    for vector in PERMISSIONS_TEST_EXCEPTION_VECTORS:
        garbage, should_raise = vector
        if should_raise:
            with pytest.raises(ValueError):
                FilePermissions(garbage)
        else:
            # shouldn't raise an exception
            FilePermissions(garbage)


def test_multiple_directories():
    dirs = FileListing(context_wrap(MULTIPLE_DIRECTORIES))
    assert '/etc/sysconfig' in dirs
    assert 'cbq' in dirs.dirs_of('/etc/sysconfig')
    # drwxr-xr-x.  2 0 0   41 Jul  6 23:32 cbq
    obj = FilePermissions.from_dict(dirs.path_entry('/etc/sysconfig/cbq'))
    assert hasattr(obj, 'name')
    assert obj.name == 'cbq'
    assert obj.perms_owner == 'rwx'
    assert obj.perms_group == 'r-x'
    assert obj.perms_other == 'r-x'

    dirs = LSlHFiles(context_wrap(LS_FILE_PERMISSIONS_DOC))
    assert '/etc/sysconfig' not in dirs
    assert '/etc/redhat-release' in dirs
    obj = dirs['/etc/redhat-release']
    assert obj.perms_owner == 'rw-'
    assert obj.perms_group == 'r--'
    assert obj.perms_other == 'r--'
    assert obj.size == 46
    assert dirs.error_lines == [
        "/bin/ls: cannot access /boot/grub/grub.conf: No such file or directory",
        "/bin/ls: cannot access fstab_mounted.devices: No such file or directory",
        "/bin/ls: cannot access pvs.devices: No such file or directory"
    ]

    assert '/dev/vda1' in dirs
    obj2 = dirs['/dev/vda1']
    assert obj2.perms_owner == 'rw-'
    assert obj2.perms_group == 'rw-'
    assert obj2.perms_other == '---'
    assert obj2.type == 'b'
    assert obj2.major == 252
    assert obj2.minor == 1



from __future__ import print_function
import doctest
import pytest

from insights.core.exceptions import ParseException, SkipComponent
from insights.parsers import lvm
from insights.tests import context_wrap
from insights.tests.parsers.lvm_test_data import LVMCONFIG, LVMCONFIG2, LVMCONFIG3

WARNINGS_CONTENT = """
WARNING
valid data 1
Checksum Error
valid data 2
  Failed to write
  Attempt To Close Device
valid data 3
""".strip()

WARNINGS_FOUND = """
WARNING
Checksum Error
  Failed to write
  Attempt To Close Device
""".strip()

VGSHEADING_CONTENT = """
Configuration setting "activation/thin_check_executable" unknown.
Configuration setting "activation/thin_check_options" unknown.
Configuration setting "activation/thin_check_executable" unknown.
Configuration setting "activation/thin_check_options" unknown.
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
  Using volume group(s) on command line.
Found duplicate PV qJMs5CEKY89qzq56E2vVeBvxzJGw2sA1: using /dev/mapper/mpathbp2 not /dev/cciss/c0d1p2
Using duplicate PV /dev/mapper/mpathbp2 from subsystem DM, ignoring /dev/cciss/c0d1p2
WARNING: Inconsistent metadata found for VG vgshared - updating to use version 26x
  Archiving volume group "vgshared" metadata (seqno 26).
  Archiving volume group "vgshared" metadata (seqno 27).
  Creating volume group backup "/etc/lvm/backup/vgshared" (seqno 27).
VG       Attr   Ext     #PV #LV #SN VSize   VFree   VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
vg00     wz--n- 128.00m   1   7   0 279.00g   5.00g MiuKdK-ruw4-UG1i-0uLE-4dfB-bk2v-uxte2W              1   506.00k  1020.00k        1
vg00alt  wz--n-   4.00m   1   5   0 279.11g 265.11g LNzjbn-HU4C-WFCM-cRZ6-81ns-4rke-X8N0DA              1   507.00k  1020.00k        1
vg01     wz--n-   4.00m   2   1   0 199.99g 100.99g 7O9ePI-M1Kp-H9VH-Lxt3-pS1Z-nLQB-P1IIO3              2   508.00k  1020.00k        2
vgshared wz--nc   4.00m   6   1   0   1.76t 899.98g fFYYw3-Ns8Q-akPI-rrFU-YYn5-nKkk-UW67qO              6   507.00k  1020.00k        6
  Reloading config files
  Wiping internal VG cache
Configuration setting "activation/thin_check_executable" unknown.
Configuration setting "activation/thin_check_options" unknown.
""".strip()

VGSHEADING_CONTENT_DOC = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
Using volume group(s) on command line.
VG            Attr   Ext   #PV #LV #SN VSize   VFree    VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
DATA_OTM_VG   wz--n- 4.00m   6   1   0   2.05t 1020.00m xK6HXk-xl2O-cqW5-2izb-LI9M-4fV0-dAzfcc              6   507.00k  1020.00k        6
ITM_VG        wz--n- 4.00m   1   1   0  16.00g    4.00m nws5dd-INe6-1db6-9U1N-F0G3-S1z2-5XTdO4              1   508.00k  1020.00k        1
ORABIN_OTM_VG wz--n- 4.00m   2   3   0 190.00g       0  hfJwg8-hset-YgUY-X6NJ-gkWE-EunZ-KuCXGP              2   507.50k  1020.00k        2
REDO_OTM_VG   wz--n- 4.00m   1   3   0  50.00g       0  Q2YtGy-CWKU-sEYj-mqHk-rbdP-Hzup-wi8jsf              1   507.50k  1020.00k        1
SWAP_OTM_VG   wz--n- 4.00m   1   1   0  24.00g    8.00g hAerzZ-U8QU-ICkc-xxCj-N2Ny-rWzq-pmTpWJ              1   508.00k  1020.00k        1
rootvg        wz--n- 4.00m   1   6   0  19.51g    1.95g p4tLLb-ikeo-Ankk-2xJ6-iHYf-D4E6-KFCFvr              1   506.50k  1020.00k        1
Reloading config files
Wiping internal VG cache
""".strip()

LVS_HAEADING_OUTUPT = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
LV          VG      Attr       LSize  Pool Origin Data%  Meta%  Move Log Cpy%Sync Convert LV Tags Devices
lv_app      vg_root -wi-ao---- 71.63g                                                             /dev/sda2(7136)
lv_home     vg_root -wi-ao----  2.00g                                                             /dev/sda2(2272)
lv_opt      vg_root -wi-ao----  5.00g                                                             /dev/sda2(2784)
lv_root     vg_root -wi-ao----  5.00g                                                             /dev/sda2(0)
lv_tmp      vg_root -wi-ao----  1.00g                                                             /dev/sda2(4064)
lv_usr      vg_root -wi-ao----  5.00g                                                             /dev/sda2(4320)
lv_usrlocal vg_root -wi-ao----  1.00g                                                             /dev/sda2(5600)
lv_var      vg_root -wi-ao----  5.00g                                                             /dev/sda2(5856)
swap        vg_root -wi-ao----  3.88g                                                             /dev/sda2(1280)
""".strip()

CONTENT_WITH_EXTRA_LOCK_TIPS = """
WARNING: locking_type (0) is deprecated, using --nolocking.
WARNING: File locking is disabled.
Reading VG shared_vg1 without a lock.
LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='V1AMDw-e9x6-AdCq-p3Hr-6cYn-otra-ducLAc'|LVM2_VG_NAME='rhel'|LVM2_VG_ATTR='wz--n-'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED=''|LVM2_VG_SIZE='<29.00g'|LVM2_VG_FREE='0 '|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE=''|LVM2_VG_LOCK_ARGS=''|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='7423'|LVM2_VG_FREE_COUNT='0'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='2'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='3'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='507.50k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
"""

LVM_CONF = """
locking_type = 1
#locking_type = 2
# volume_list = [ "vg1", "vg2/lvol1", "@tag1", "@*" ]
volume_list = [ "vg2", "vg3/lvol3", "@tag2", "@*" ]
# filter = [ "a|loop|", "r|/dev/hdc|", "a|/dev/ide|", "r|.*|" ]

filter = [ "r/sda[0-9]*$/",  "a/sd.*/" ]
filter = [ "a/sda[0-9]*$/",  "r/sd.*/" ]
shell {
    history_size = 100
}
""".strip()

PVS_HEADINGS_OUTPUT = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
Scanning all devices to update lvmetad.
No PV label found on /dev/loop0.
No PV label found on /dev/loop1.
No PV label found on /dev/sda1.
No PV label found on /dev/fedora/root.
No PV label found on /dev/sda2.
No PV label found on /dev/fedora/swap.
No PV label found on /dev/fedora/home.
No PV label found on /dev/mapper/docker-253:1-2361272-pool.
Wiping internal VG cache
Wiping cache of LVM-capable devices
PV                                                    VG     Fmt  Attr PSize   PFree DevSize PV UUID                                PMdaFree  PMdaSize  #PMda #PMdaUse PE
/dev/fedora/home                                                  ---       0     0  418.75g                                               0         0      0        0      0
/dev/fedora/root                                                  ---       0     0   50.00g                                               0         0      0        0      0
/dev/fedora/swap                                                  ---       0     0    7.69g                                               0         0      0        0      0
/dev/loop0                                                        ---       0     0  100.00g                                               0         0      0        0      0
/dev/loop1                                                        ---       0     0    2.00g                                               0         0      0        0      0
/dev/mapper/docker-253:1-2361272-pool                             ---       0     0  100.00g                                               0         0      0        0      0
/dev/mapper/luks-7430952e-7101-4716-9b46-786ce4684f8d fedora lvm2 a--  476.45g 4.00m 476.45g FPLCRf-d918-LVL7-6e3d-n3ED-aiZv-EesuzY        0   1020.00k     1        1 121970
/dev/sda1                                                         ---       0     0  500.00m                                               0         0      0        0      0
/dev/sda2                                                         ---       0     0  476.45g                                               0         0      0        0      0
Reloading config files
Wiping internal VG cache
""".strip()

VGS_OUTPUT = """
WARNING: Locking disabled. Be careful! This could corrupt your metadata.
Using volume group(s) on command line.
VG            Attr   Ext   #PV #LV #SN VSize   VFree    VG UUID                                VProfile #VMda VMdaFree  VMdaSize  #VMdaUse VG Tags
DATA_OTM_VG   wz--n- 4.00m   6   1   0   2.05t 1020.00m xK6HXk-xl2O-cqW5-2izb-LI9M-4fV0-dAzfcc              6   507.00k  1020.00k        6
ITM_VG        wz--n- 4.00m   1   1   0  16.00g    4.00m nws5dd-INe6-1db6-9U1N-F0G3-S1z2-5XTdO4              1   508.00k  1020.00k        1
ORABIN_OTM_VG wz--n- 4.00m   2   3   0 190.00g       0  hfJwg8-hset-YgUY-X6NJ-gkWE-EunZ-KuCXGP              2   507.50k  1020.00k        2
REDO_OTM_VG   wz--n- 4.00m   1   3   0  50.00g       0  Q2YtGy-CWKU-sEYj-mqHk-rbdP-Hzup-wi8jsf              1   507.50k  1020.00k        1
SWAP_OTM_VG   wz--n- 4.00m   1   1   0  24.00g    8.00g hAerzZ-U8QU-ICkc-xxCj-N2Ny-rWzq-pmTpWJ              1   508.00k  1020.00k        1
rootvg        wz--n- 4.00m   1   6   0  19.51g    1.95g p4tLLb-ikeo-Ankk-2xJ6-iHYf-D4E6-KFCFvr              1   506.50k  1020.00k        1
Reloading config files
Wiping internal VG cache
""".strip()

SYSTEM_DEVICES1 = """
# LVM uses devices listed in this file.
# Created by LVM command lvmdevices pid 2631 at Fri May 27 07:37:11 2022
VERSION=1.1.2
IDTYPE=devname IDNAME=/dev/vda2 DEVNAME=/dev/vda2 PVID=phl0clFbAokp9UXqbIgI5YYQxuTIJVkD PART=2
""".strip()

SYSTEM_DEVICES2 = """
# LVM uses devices listed in this file.
# Created by LVM command lvmdevices pid 2631 at Fri May 27 07:37:11 2022
VERSION=1.1.2
IDTYPE=sys_wwid IDNAME=/dev/vda2 DEVNAME=/dev/vda2 PVID=phl0clFbAokp9UXqbIgI5YYQxuTIJVkD PART=2
IDTYPE=sys_serial IDNAME=/dev/vda1 DEVNAME=/dev/vda1 PVID=phl0clFbAokp9UXqbIgI5YYQdeTIJVkD PART=1
""".strip()

SYSTEM_DEVICES3 = """
# LVM uses devices listed in this file.
# Created by LVM command lvmdevices pid 2631 at Fri May 27 07:37:11 2022
VERSION=1.1.2
""".strip()

LVM_FULLREPORT = """
{
    "report": [
        {
          "vg": [{"vg_fmt":"lvm2", "vg_uuid":"zclHZK-pMKd-fSIC-8TRX-xitj-RxY1-vwCRAw", "vg_name":"rhel", "vg_attr":"wz--n-"}],
          "pv": [{"pv_fmt":"lvm2", "pv_uuid":"LU7am7-ACyz-cPlb-byCA-j4IM-9FK8-iS1WA4", "dev_size":"<9.00g", "pv_name":"/dev/vda2"}],
          "lv": [{"lv_uuid":"kG4B7g-Ggar-uwtO-EbEX-gBX0-EGO2-Y0VFfU", "lv_name":"root", "lv_full_name":"rhel/root", "lv_path":"/dev/rhel/root"}],
          "pvseg": [{"pvseg_start":"0", "pvseg_size":"256", "pv_uuid":"LU7am7-ACyz-cPlb-byCA-j4IM-9FK8-iS1WA4", "lv_uuid":"8p6Tu0-OGa9-rbjg-5V22-D1dX-lNw7-Jed3m7"}],
          "seg": [{"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1"}]
        },
        {
          "vg": [{"vg_fmt":"lvm2", "vg_uuid":"1rLhOL-Mqkc-gi9I-Yspd-QZ3W-ZWN6-5viZiT", "vg_name":"vg1", "vg_attr":"wz--n-"}],
          "pv": [{"pv_fmt":"lvm2", "pv_uuid":"WwYITo-ng46-4Ufw-DqKx-0ZQA-LQeh-Vnga1N", "dev_size":"5.00g", "pv_name":"/dev/vdb"}],
          "lv": [{"lv_uuid":"tZdwGc-xGEg-VMQU-WIgE-k8kP-Udeo-hgLofA", "lv_name":"lvraid1", "lv_full_name":"vg1/lvraid1", "lv_path":"/dev/vg1/lvraid1"}],
          "pvseg": [{"pvseg_start":"0", "pvseg_size":"1", "pv_uuid":"D39Poc-PupR-sYD5-bUeJ-2aYO-YWdP-XAInA5", "lv_uuid":"3YHTyV-HXoc-wRRH-D1Jw-NNIs-m2g8-cOvgGS"}],
          "seg": [{"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1"}]
        }
    ]
}
""".strip()

LVM_FULLREPORT_WARNING = """
WARNING: locking_type(0) is deprecated, using - -nolocking.
{
    "report": [
        {
          "vg": [{"vg_fmt":"lvm2", "vg_uuid":"zclHZK-pMKd-fSIC-8TRX-xitj-RxY1-vwCRAw", "vg_name":"rhel", "vg_attr":"wz--n-"}],
          "pv": [{"pv_fmt":"lvm2", "pv_uuid":"LU7am7-ACyz-cPlb-byCA-j4IM-9FK8-iS1WA4", "dev_size":"<9.00g", "pv_name":"/dev/vda2"}],
          "lv": [{"lv_uuid":"kG4B7g-Ggar-uwtO-EbEX-gBX0-EGO2-Y0VFfU", "lv_name":"root", "lv_full_name":"rhel/root", "lv_path":"/dev/rhel/root"}],
          "pvseg": [{"pvseg_start":"0", "pvseg_size":"256", "pv_uuid":"LU7am7-ACyz-cPlb-byCA-j4IM-9FK8-iS1WA4", "lv_uuid":"8p6Tu0-OGa9-rbjg-5V22-D1dX-lNw7-Jed3m7"}],
          "seg": [{"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1"}]
        },
        {
          "vg": [{"vg_fmt":"lvm2", "vg_uuid":"1rLhOL-Mqkc-gi9I-Yspd-QZ3W-ZWN6-5viZiT", "vg_name":"vg1", "vg_attr":"wz--n-"}],
          "pv": [{"pv_fmt":"lvm2", "pv_uuid":"WwYITo-ng46-4Ufw-DqKx-0ZQA-LQeh-Vnga1N", "dev_size":"5.00g", "pv_name":"/dev/vdb"}],
          "lv": [{"lv_uuid":"tZdwGc-xGEg-VMQU-WIgE-k8kP-Udeo-hgLofA", "lv_name":"lvraid1", "lv_full_name":"vg1/lvraid1", "lv_path":"/dev/vg1/lvraid1"}],
          "pvseg": [{"pvseg_start":"0", "pvseg_size":"1", "pv_uuid":"D39Poc-PupR-sYD5-bUeJ-2aYO-YWdP-XAInA5", "lv_uuid":"3YHTyV-HXoc-wRRH-D1Jw-NNIs-m2g8-cOvgGS"}],
          "seg": [{"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1"}]
        }
    ]
}
""".strip()

LVM_FULLREPORT_WARNING_IN_JSON_MIDDLE = """
  WARNING: locking_type(0) is deprecated, using - -nolocking.
  {
      "report": [
          {
          "vg": [{"vg_fmt":"lvm2", "vg_uuid":"zclHZK-pMKd-fSIC-8TRX-xitj-RxY1-vwCRAw", "vg_name":"rhel", "vg_attr":"wz--n-"}],
          "pv": [{"pv_fmt":"lvm2", "pv_uuid":"LU7am7-ACyz-cPlb-byCA-j4IM-9FK8-iS1WA4", "dev_size":"<9.00g", "pv_name":"/dev/vda2"}],
          "lv": [{"lv_uuid":"kG4B7g-Ggar-uwtO-EbEX-gBX0-EGO2-Y0VFfU", "lv_name":"root", "lv_full_name":"rhel/root", "lv_path":"/dev/rhel/root"}],
          "pvseg": [{"pvseg_start":"0", "pvseg_size":"256", "pv_uuid":"LU7am7-ACyz-cPlb-byCA-j4IM-9FK8-iS1WA4", "lv_uuid":"8p6Tu0-OGa9-rbjg-5V22-D1dX-lNw7-Jed3m7"}],
          "seg": [{"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1"}]
          }
          ,
  WARNING: PV /dev/sdc in VG oraredo is using an old PV header, modify the VG to update.
          {
          "vg": [{"vg_fmt":"lvm2", "vg_uuid":"1rLhOL-Mqkc-gi9I-Yspd-QZ3W-ZWN6-5viZiT", "vg_name":"vg1", "vg_attr":"wz--n-"}],
          "pv": [{"pv_fmt":"lvm2", "pv_uuid":"WwYITo-ng46-4Ufw-DqKx-0ZQA-LQeh-Vnga1N", "dev_size":"5.00g", "pv_name":"/dev/vdb"}],
          "lv": [{"lv_uuid":"tZdwGc-xGEg-VMQU-WIgE-k8kP-Udeo-hgLofA", "lv_name":"lvraid1", "lv_full_name":"vg1/lvraid1", "lv_path":"/dev/vg1/lvraid1"}],
          "pvseg": [{"pvseg_start":"0", "pvseg_size":"1", "pv_uuid":"D39Poc-PupR-sYD5-bUeJ-2aYO-YWdP-XAInA5", "lv_uuid":"3YHTyV-HXoc-wRRH-D1Jw-NNIs-m2g8-cOvgGS"}],
          "seg": [{"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1"}]
          }
      ]
  }
""".strip()

LVM_FULLREPORT_SPECIAL_KEYWORDS = """
  WARNING: locking_type(0) is deprecated, using - -nolocking.
  {
    "report": [
  WARNING: PV /dev/sdd in VG vgfo is using an old PV header, modify the VG to update.
      {
          "vg": [
              {"vg_fmt":"lvm2", "vg_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc4", "vg_name":"vgfo", "vg_attr":"wz--n-", "vg_permissions":"writeable", "vg_extendable":"extendable", "vg_exported":"", "vg_autoactivation":"enabled", "vg_partial":"", "vg_allocation_policy":"normal", "vg_clustered":"", "vg_shared":"", "vg_size":"1.99g", "vg_free":"<1.47g", "vg_sysid":"", "vg_systemid":"", "vg_lock_type":"", "vg_lock_args":"", "vg_extent_size":"4.00m", "vg_extent_count":"510", "vg_free_count":"376", "max_lv":"0", "max_pv":"0", "pv_count":"2", "vg_missing_pv_count":"0", "lv_count":"3", "snap_count":"0", "vg_seqno":"65", "vg_tags":"TTT", "vg_profile":"", "vg_mda_count":"2", "vg_mda_used_count":"2", "vg_mda_free":"506.00k", "vg_mda_size":"1020.00k", "vg_mda_copies":"unmanaged"}
          ]
          ,
          "pv": [
              {"pv_fmt":"lvm2", "pv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc5", "dev_size":"1.00g", "pv_name":"/dev/sdd", "pv_major":"8", "pv_minor":"48", "pv_mda_free":"506.00k", "pv_mda_size":"1020.00k", "pv_ext_vsn":"2", "pe_start":"1.00m", "pv_size":"1020.00m", "pv_free":"752.00m", "pv_used":"268.00m", "pv_attr":"a--", "pv_allocatable":"allocatable", "pv_exported":"", "pv_missing":"", "pv_pe_count":"255", "pv_pe_alloc_count":"67", "pv_tags":"", "pv_mda_count":"1", "pv_mda_used_count":"1", "pv_ba_start":"0 ", "pv_ba_size":"0 ", "pv_in_use":"used", "pv_duplicate":"", "pv_device_id":"", "pv_device_id_type":""},
              {"pv_fmt":"lvm2", "pv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc6", "dev_size":"1.00g", "pv_name":"/dev/sde", "pv_major":"8", "pv_minor":"64", "pv_mda_free":"506.00k", "pv_mda_size":"1020.00k", "pv_ext_vsn":"2", "pe_start":"1.00m", "pv_size":"1020.00m", "pv_free":"752.00m", "pv_used":"268.00m", "pv_attr":"a--", "pv_allocatable":"allocatable", "pv_exported":"", "pv_missing":"", "pv_pe_count":"255", "pv_pe_alloc_count":"67", "pv_tags":"", "pv_mda_count":"1", "pv_mda_used_count":"1", "pv_ba_start":"0 ", "pv_ba_size":"0 ", "pv_in_use":"used", "pv_duplicate":"", "pv_device_id":"", "pv_device_id_type":""}
          ]
          ,
          "lv": [
              {"lv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc7", "lv_name":"lv_d", "lv_full_name":"vgfo/lv_d", "lv_path":"/dev/vgfo/lv_d", "lv_dm_path":"/dev/mapper/vgfo-lv_d", "lv_parent":"", "lv_layout":"linear", "lv_role":"public", "lv_initial_image_sync":"", "lv_image_synced":"", "lv_merging":"", "lv_converting":"", "lv_allocation_policy":"inherit", "lv_allocation_locked":"", "lv_fixed_minor":"", "lv_skip_activation":"", "lv_autoactivation":"enabled", "lv_when_full":"", "lv_active":"", "lv_active_locally":"", "lv_active_remotely":"", "lv_active_exclusively":"", "lv_major":"-1", "lv_minor":"-1", "lv_read_ahead":"auto", "lv_size":"4.00m", "lv_metadata_size":"", "seg_count":"1", "origin":"", "origin_uuid":"", "origin_size":"", "lv_ancestors":"", "lv_full_ancestors":"", "lv_descendants":"", "lv_full_descendants":"", "raid_mismatch_count":"", "raid_sync_action":"", "raid_write_behind":"", "raid_min_recovery_rate":"", "raid_max_recovery_rate":"", "raidintegritymode":"", "raidintegrityblocksize":"-1", "integritymismatches":"", "move_pv":"", "move_pv_uuid":"", "convert_lv":"", "convert_lv_uuid":"", "mirror_log":"", "mirror_log_uuid":"", "data_lv":"", "data_lv_uuid":"", "metadata_lv":"", "metadata_lv_uuid":"", "pool_lv":"", "pool_lv_uuid":"", "lv_tags":"", "lv_profile":"", "lv_lockargs":"", "lv_time":"2023-07-06 14:08:52 -0500", "lv_time_removed":"", "lv_host":"r8n0", "lv_modules":"", "lv_historical":"", "writecache_block_size":"-1", "lv_kernel_major":"-1", "lv_kernel_minor":"-1", "lv_kernel_read_ahead":"-1", "lv_permissions":"unknown", "lv_suspended":"unknown", "lv_live_table":"unknown", "lv_inactive_table":"unknown", "lv_device_open":"unknown", "data_percent":"", "snap_percent":"", "metadata_percent":"", "copy_percent":"", "sync_percent":"", "cache_total_blocks":"", "cache_used_blocks":"", "cache_dirty_blocks":"", "cache_read_hits":"", "cache_read_misses":"", "cache_write_hits":"", "cache_write_misses":"", "kernel_cache_settings":"", "kernel_cache_policy":"", "kernel_metadata_format":"", "lv_health_status":"", "kernel_discards":"", "lv_check_needed":"unknown", "lv_merge_failed":"unknown", "lv_snapshot_invalid":"unknown", "vdo_operating_mode":"", "vdo_compression_state":"", "vdo_index_state":"", "vdo_used_size":"", "vdo_saving_percent":"", "writecache_total_blocks":"", "writecache_free_blocks":"", "writecache_writeback_blocks":"", "writecache_error":"", "lv_attr":"-wi-------"},
              {"lv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc8", "lv_name":"lv_e", "lv_full_name":"vgfo/lv_e", "lv_path":"/dev/vgfo/lv_e", "lv_dm_path":"/dev/mapper/vgfo-lv_e", "lv_parent":"", "lv_layout":"linear", "lv_role":"public", "lv_initial_image_sync":"", "lv_image_synced":"", "lv_merging":"", "lv_converting":"", "lv_allocation_policy":"inherit", "lv_allocation_locked":"", "lv_fixed_minor":"", "lv_skip_activation":"", "lv_autoactivation":"enabled", "lv_when_full":"", "lv_active":"", "lv_active_locally":"", "lv_active_remotely":"", "lv_active_exclusively":"", "lv_major":"-1", "lv_minor":"-1", "lv_read_ahead":"auto", "lv_size":"4.00m", "lv_metadata_size":"", "seg_count":"1", "origin":"", "origin_uuid":"", "origin_size":"", "lv_ancestors":"", "lv_full_ancestors":"", "lv_descendants":"", "lv_full_descendants":"", "raid_mismatch_count":"", "raid_sync_action":"", "raid_write_behind":"", "raid_min_recovery_rate":"", "raid_max_recovery_rate":"", "raidintegritymode":"", "raidintegrityblocksize":"-1", "integritymismatches":"", "move_pv":"", "move_pv_uuid":"", "convert_lv":"", "convert_lv_uuid":"", "mirror_log":"", "mirror_log_uuid":"", "data_lv":"", "data_lv_uuid":"", "metadata_lv":"", "metadata_lv_uuid":"", "pool_lv":"", "pool_lv_uuid":"", "lv_tags":"", "lv_profile":"", "lv_lockargs":"", "lv_time":"2023-07-06 14:08:43 -0500", "lv_time_removed":"", "lv_host":"r8n0", "lv_modules":"", "lv_historical":"", "writecache_block_size":"-1", "lv_kernel_major":"-1", "lv_kernel_minor":"-1", "lv_kernel_read_ahead":"-1", "lv_permissions":"unknown", "lv_suspended":"unknown", "lv_live_table":"unknown", "lv_inactive_table":"unknown", "lv_device_open":"unknown", "data_percent":"", "snap_percent":"", "metadata_percent":"", "copy_percent":"", "sync_percent":"", "cache_total_blocks":"", "cache_used_blocks":"", "cache_dirty_blocks":"", "cache_read_hits":"", "cache_read_misses":"", "cache_write_hits":"", "cache_write_misses":"", "kernel_cache_settings":"", "kernel_cache_policy":"", "kernel_metadata_format":"", "lv_health_status":"", "kernel_discards":"", "lv_check_needed":"unknown", "lv_merge_failed":"unknown", "lv_snapshot_invalid":"unknown", "vdo_operating_mode":"", "vdo_compression_state":"", "vdo_index_state":"", "vdo_used_size":"", "vdo_saving_percent":"", "writecache_total_blocks":"", "writecache_free_blocks":"", "writecache_writeback_blocks":"", "writecache_error":"", "lv_attr":"-wi-------"}
          ]
          ,
          "pvseg": [
              {"pvseg_start":"0", "pvseg_size":"1", "pv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc5", "lv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc7"},
              {"pvseg_start":"1", "pvseg_size":"64", "pv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc6", "lv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc8"}
          ]
          ,
          "seg": [
              {"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1", "data_offset":"", "new_data_offset":"", "parity_chunks":"", "stripe_size":"0 ", "region_size":"0 ", "chunk_size":"0 ", "thin_count":"", "discards":"", "cache_metadata_format":"", "cache_mode":"", "zero":"unknown", "transaction_id":"", "thin_id":"", "seg_start":"0 ", "seg_start_pe":"0", "seg_size":"4.00m", "seg_size_pe":"1", "seg_tags":"", "seg_pe_ranges":"/dev/sde:65-65", "seg_le_ranges":"/dev/sde:65-65", "seg_metadata_le_ranges":"", "devices":"/dev/sde(65)", "metadata_devices":"", "seg_monitor":"", "cache_policy":"", "cache_settings":"", "vdo_compression":"", "vdo_deduplication":"", "vdo_use_metadata_hints":"", "vdo_minimum_io_size":"", "vdo_block_map_cache_size":"", "vdo_block_map_era_length":"", "vdo_use_sparse_index":"", "vdo_index_memory_size":"", "vdo_slab_size":"", "vdo_ack_threads":"", "vdo_bio_threads":"", "vdo_bio_rotation":"", "vdo_cpu_threads":"", "vdo_hash_zone_threads":"", "vdo_logical_threads":"", "vdo_physical_threads":"", "vdo_max_discard":"", "vdo_write_policy":"", "vdo_header_size":"", "lv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc9"},
              {"segtype":"linear", "stripes":"1", "data_stripes":"1", "reshape_len":"", "reshape_len_le":"", "data_copies":"1", "data_offset":"", "new_data_offset":"", "parity_chunks":"", "stripe_size":"0 ", "region_size":"0 ", "chunk_size":"0 ", "thin_count":"", "discards":"", "cache_metadata_format":"", "cache_mode":"", "zero":"unknown", "transaction_id":"", "thin_id":"", "seg_start":"0 ", "seg_start_pe":"0", "seg_size":"4.00m", "seg_size_pe":"1", "seg_tags":"", "seg_pe_ranges":"/dev/sde:0-0", "seg_le_ranges":"/dev/sde:0-0", "seg_metadata_le_ranges":"", "devices":"/dev/sde(0)", "metadata_devices":"", "seg_monitor":"", "cache_policy":"", "cache_settings":"", "vdo_compression":"", "vdo_deduplication":"", "vdo_use_metadata_hints":"", "vdo_minimum_io_size":"", "vdo_block_map_cache_size":"", "vdo_block_map_era_length":"", "vdo_use_sparse_index":"", "vdo_index_memory_size":"", "vdo_slab_size":"", "vdo_ack_threads":"", "vdo_bio_threads":"", "vdo_bio_rotation":"", "vdo_cpu_threads":"", "vdo_hash_zone_threads":"", "vdo_logical_threads":"", "vdo_physical_threads":"", "vdo_max_discard":"", "vdo_write_policy":"", "vdo_header_size":"", "lv_uuid":"43522d-03cc-11f0-9164-62c3-585c-904cc0"}
          ]
      }
    ]
}""".strip()

LVM_FULLREPORT_EMPTY = """
{
    "report": [
    ]
}
""".strip()

LVM_FULLREPORT_NO_JSON = """
WARNING: locking_type(0) is deprecated, using - -nolocking.
""".strip()

VGS_WITH_FOREIGN_STRING = """
  LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='9nEuop-JraN-SMv7-DxpS-J1wC-NZvV-fzPtuk'|LVM2_VG_NAME='cluster_vg'|LVM2_VG_ATTR='wz--n-'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION=''|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED=''|LVM2_VG_SIZE='<20.00g'|LVM2_VG_FREE='0 '|LVM2_VG_SYSID='node1.redhat.com'|LVM2_VG_SYSTEMID='node1.redhat.com'|LVM2_VG_LOCK_TYPE=''|LVM2_VG_LOCK_ARGS=''|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='5119'|LVM2_VG_FREE_COUNT='0'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='1'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='5'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='507.50k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
  LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='SHRR9U-WPAB-m25d-8zz7-If1j-lXoX-GpA0lI'|LVM2_VG_NAME='rhel_rhel9'|LVM2_VG_ATTR='wz--n-'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED=''|LVM2_VG_SIZE='<127.00g'|LVM2_VG_FREE='54.96g'|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE=''|LVM2_VG_LOCK_ARGS=''|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='32511'|LVM2_VG_FREE_COUNT='14070'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='2'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='7'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='507.50k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
"""

VGS_WITH_SHARED_STRING = """
  WARNING: File locking is disabled.
  LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='ms3Po2-XPJP-QLk7-Dh29-JNvQ-9bSN-JavZVj'|LVM2_VG_NAME='cluster_vg'|LVM2_VG_ATTR='wz--ns'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED='shared'|LVM2_VG_SIZE='<10.00g'|LVM2_VG_FREE='0 '|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE='dlm'|LVM2_VG_LOCK_ARGS='1.0.0:hh_cluster'|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='2559'|LVM2_VG_FREE_COUNT='0'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='1'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='3'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='508.00k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
  LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='Y49xze-fTp1-PypT-vIqE-LlSc-tkyH-uhgdZv'|LVM2_VG_NAME='rhel_rhel8'|LVM2_VG_ATTR='wz--n-'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED=''|LVM2_VG_SIZE='<127.00g'|LVM2_VG_FREE='<54.94g'|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE=''|LVM2_VG_LOCK_ARGS=''|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='32511'|LVM2_VG_FREE_COUNT='14064'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='2'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='7'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='507.50k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
"""

VGS_WITH_CLUSTERED_STRING = """
  WARNING: File locking is disabled.
  LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='ms3Po2-XPJP-QLk7-Dh29-JNvQ-9bSN-JavZVj'|LVM2_VG_NAME='cluster_vg'|LVM2_VG_ATTR='wz--nc'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED='shared'|LVM2_VG_SIZE='<10.00g'|LVM2_VG_FREE='0 '|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE='dlm'|LVM2_VG_LOCK_ARGS='1.0.0:hh_cluster'|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='2559'|LVM2_VG_FREE_COUNT='0'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='1'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='3'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='508.00k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
  LVM2_VG_FMT='lvm2'|LVM2_VG_UUID='Y49xze-fTp1-PypT-vIqE-LlSc-tkyH-uhgdZv'|LVM2_VG_NAME='rhel_rhel8'|LVM2_VG_ATTR='wz--n-'|LVM2_VG_PERMISSIONS='writeable'|LVM2_VG_EXTENDABLE='extendable'|LVM2_VG_EXPORTED=''|LVM2_VG_AUTOACTIVATION='enabled'|LVM2_VG_PARTIAL=''|LVM2_VG_ALLOCATION_POLICY='normal'|LVM2_VG_CLUSTERED=''|LVM2_VG_SHARED=''|LVM2_VG_SIZE='<127.00g'|LVM2_VG_FREE='<54.94g'|LVM2_VG_SYSID=''|LVM2_VG_SYSTEMID=''|LVM2_VG_LOCK_TYPE=''|LVM2_VG_LOCK_ARGS=''|LVM2_VG_EXTENT_SIZE='4.00m'|LVM2_VG_EXTENT_COUNT='32511'|LVM2_VG_FREE_COUNT='14064'|LVM2_MAX_LV='0'|LVM2_MAX_PV='0'|LVM2_PV_COUNT='1'|LVM2_VG_MISSING_PV_COUNT='0'|LVM2_LV_COUNT='2'|LVM2_SNAP_COUNT='0'|LVM2_VG_SEQNO='7'|LVM2_VG_TAGS=''|LVM2_VG_PROFILE=''|LVM2_VG_MDA_COUNT='1'|LVM2_VG_MDA_USED_COUNT='1'|LVM2_VG_MDA_FREE='507.50k'|LVM2_VG_MDA_SIZE='1020.00k'|LVM2_VG_MDA_COPIES='unmanaged'
"""


def test_find_warnings():
    data = [l for l in lvm.find_warnings(WARNINGS_CONTENT.splitlines())]
    assert len(data) == len(WARNINGS_FOUND.splitlines())
    assert data == [0, 2, 4, 5]  # The WARNINGS_FOUND lines index in WARNINGS_CONTENT


def compare_partial_dicts(result, expected):
    """
    Make sure all the keys in expected are matched by keys in result, and
    that the values stored in those keys match.  Result can contain more
    items than expected - those are ignored.

    Used in the test_lvs, test_pvs and test_vgs tests.
    """
    # return all(result[k] == expected[k] for k in expected.keys())
    mismatches = 0
    for k in expected.keys():
        if not result[k] == expected[k]:
            print("Failed for key {k}, {r} != {e}".format(k=k, r=result[k], e=expected[k]))
            mismatches += 1
    return mismatches == 0


def test_lvmconfig():
    p = lvm.LvmConfig(context_wrap(LVMCONFIG))
    assert p.data["dmeventd"]["raid_library"] == "libdevmapper-event-lvm2raid.so"
    assert p.data["global"]["thin_check_options"] == ["-q", "--clear-needs-check-flag"]


def test_lvmconfig_2():
    p = lvm.LvmConfig(context_wrap(LVMCONFIG2))
    assert p.data['global']['umask'] == '077'


def test_lvmconfig_exception():
    with pytest.raises(ParseException):
        lvm.LvmConfig(context_wrap(LVMCONFIG3))


def test_vgsheading_warnings():
    result = lvm.VgsHeadings(context_wrap(VGSHEADING_CONTENT))
    assert len(result.warnings) == 6
    assert 'Configuration setting "activation/thin_check_executable" unknown.' in result.warnings
    assert (
        'WARNING: Locking disabled. Be careful! This could corrupt your metadata.'
        in result.warnings
    )


def test_vgs_with_extra_tips():
    result = lvm.Vgs(context_wrap(CONTENT_WITH_EXTRA_LOCK_TIPS))
    assert result is not None
    assert len(result.data.get('warnings')) == 3
    assert 'Reading VG shared_vg1 without a lock.' in result.data.get('warnings')
    assert len(result.data.get('content')) == 1


def test_system_devices():
    devices = lvm.LvmSystemDevices(context_wrap(SYSTEM_DEVICES2))
    assert len(devices) == 2
    assert '/dev/vda1' in devices
    assert devices['/dev/vda1']['IDTYPE'] == 'sys_serial'
    assert '/dev/vda2' in devices
    assert devices['/dev/vda2']['IDTYPE'] == 'sys_wwid'


def test_system_devices_exception():
    with pytest.raises(SkipComponent):
        lvm.LvmSystemDevices(context_wrap(SYSTEM_DEVICES3))


def test_lvm_fullreport():
    for data in [LVM_FULLREPORT, LVM_FULLREPORT_WARNING, LVM_FULLREPORT_WARNING_IN_JSON_MIDDLE]:
        report = lvm.LvmFullReport(context_wrap(data))
        assert report is not None
        assert len(report.volume_groups) == 2
        assert 'vg1' in report.volume_groups and 'rhel' in report.volume_groups
        vg1 = report.volume_groups['vg1']
        assert set(vg1.keys()) == set(['vg', 'pv', 'lv', 'pvseg', 'seg'])
        assert vg1['vg'][0]['vg_name'] == 'vg1'
        assert vg1['pv'][0]['pv_name'] == '/dev/vdb'
        assert vg1['lv'][0]['lv_name'] == 'lvraid1'
        assert vg1['pvseg'][0] == {
            "pvseg_start": "0",
            "pvseg_size": "1",
            "pv_uuid": "D39Poc-PupR-sYD5-bUeJ-2aYO-YWdP-XAInA5",
            "lv_uuid": "3YHTyV-HXoc-wRRH-D1Jw-NNIs-m2g8-cOvgGS",
        }
        assert vg1['seg'][0]['data_stripes'] == '1'

        if data == LVM_FULLREPORT_WARNING:
            assert report.warnings == [
                LVM_FULLREPORT_WARNING.splitlines()[0],
            ]

        if data == LVM_FULLREPORT_WARNING_IN_JSON_MIDDLE:
            assert report.warnings == [
                LVM_FULLREPORT_WARNING_IN_JSON_MIDDLE.splitlines()[0],
                LVM_FULLREPORT_WARNING_IN_JSON_MIDDLE.splitlines()[11],
            ]

    report = lvm.LvmFullReport(context_wrap(LVM_FULLREPORT_SPECIAL_KEYWORDS))
    assert report is not None
    assert len(report.volume_groups) == 1
    assert 'vgfo' in report.volume_groups
    vgfo = report.volume_groups['vgfo']
    assert set(vgfo.keys()) == set(['vg', 'pv', 'lv', 'pvseg', 'seg'])
    assert vgfo['vg'][0]['vg_name'] == 'vgfo'
    assert vgfo['pv'][0]['pv_name'] == '/dev/sdd'
    assert vgfo['lv'][0]['lv_name'] == 'lv_d'
    assert vgfo['pvseg'][0]['pvseg_size'] == '1'
    assert vgfo['seg'][0]['data_stripes'] == '1'
    assert report.warnings == [
        LVM_FULLREPORT_SPECIAL_KEYWORDS.splitlines()[0],
        LVM_FULLREPORT_SPECIAL_KEYWORDS.splitlines()[3],
    ]


def test_lvm_fullreport_empty():
    with pytest.raises(SkipComponent) as e:
        lvm.LvmFullReport(context_wrap(LVM_FULLREPORT_EMPTY))
    assert "No LVM information in fullreport" in str(e)


def test_vgs_with_foreign_and_share():
    data = lvm.VgsWithForeignAndShared(context_wrap(VGS_WITH_FOREIGN_STRING))
    assert len(data) == 2
    assert not data.shared_vgs
    assert not data.clustered_vgs
    assert data[0]['VG_UUID'] == '9nEuop-JraN-SMv7-DxpS-J1wC-NZvV-fzPtuk'
    assert data[0]['LVM2_VG_SYSTEMID'] == 'node1.redhat.com'

    assert data[1]['VG_UUID'] == 'SHRR9U-WPAB-m25d-8zz7-If1j-lXoX-GpA0lI'
    assert data[1]['LVM2_VG_SYSTEMID'] == ''

    data2 = lvm.VgsWithForeignAndShared(context_wrap(VGS_WITH_SHARED_STRING))
    assert len(data2) == 2
    assert len(data2.shared_vgs) == 1
    data2.shared_vgs[0]['VG_UUID'] == 'ms3Po2-XPJP-QLk7-Dh29-JNvQ-9bSN-JavZVj'
    data2.shared_vgs[0]['VG'] == 'cluster_vg'
    data2.shared_vgs[0]['Attr'] == 'wz--ns'
    assert not data2.clustered_vgs

    data3 = lvm.VgsWithForeignAndShared(context_wrap(VGS_WITH_CLUSTERED_STRING))
    assert len(data3) == 2
    assert not data3.shared_vgs
    assert len(data3.clustered_vgs) == 1
    data3.clustered_vgs[0]['VG_UUID'] == 'ms3Po2-XPJP-QLk7-Dh29-JNvQ-9bSN-JavZVj'
    data3.clustered_vgs[0]['VG'] == 'cluster_vg'
    data3.clustered_vgs[0]['Attr'] == 'wz--nc'


def test_docs():
    env = {
        'devices': lvm.LvmSystemDevices(context_wrap(SYSTEM_DEVICES1)),
        'lvm_conf_data': lvm.LvmConf(context_wrap(LVM_CONF)),
        'pvs_data': lvm.PvsHeadings(context_wrap(PVS_HEADINGS_OUTPUT)),
        'vgs_info': lvm.VgsHeadings(context_wrap(VGSHEADING_CONTENT_DOC)),
        'lvs_info': lvm.LvsHeadings(context_wrap(LVS_HAEADING_OUTUPT)),
        'lvm_fullreport': lvm.LvmFullReport(context_wrap(LVM_FULLREPORT)),
        'vgs_all_data': lvm.VgsWithForeignAndShared(context_wrap(VGS_WITH_FOREIGN_STRING)),
    }
    failed, total = doctest.testmod(lvm, globs=env)
    assert failed == 0



"""
Filefrag - Command ``/sbin/filefrag <specified file>``
======================================================

Parser for parsing the output of command ``/sbin/filefrag <specified file>``.
"""
from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.filefrag)
class Filefrag(Parser, dict):
    """
    Provides the extent counts of the specified file by parsing the
    output of command ``filefrag <specified file>``.

    Typical content looks like::

        open: No such file or directory
        /boot/grub2/grubenv: 1 extent found

    Examples:
        >>> type(filefrag)
        <class 'insights.parsers.filefrag.Filefrag'>
        >>> filefrag['/boot/grub2/grubenv']
        1
        >>> filefrag.unparsed_lines
        []

    Raises:
        SkipComponent: if the command output is empty or missing file
    """

    def parse_content(self, content):
        self.unparsed_lines = []
        for line in content:
            if 'No such file or directory' in line:
                # Skip non-exist dirs
                continue
            try:
                file, frag = line.split(': ')
                extents = frag.strip().split()[0]
                self[file] = int(extents)
            except Exception:
                self.unparsed_lines.append(line)
        if not self and not self.unparsed_lines:
            raise SkipComponent



"""
Password Replacement
====================

"""

import logging
import re

logger = logging.getLogger(__name__)

DEFAULT_PASSWORD_REGEXS = [
    r"(password[a-zA-Z0-9_]*)(\s*\:\s*\"*\s*|\s*\"*\s*=\s*\"\s*|\s*=+\s*|\s*--md5+\s*|\s*)([a-zA-Z0-9_!@#$%^&*()+=/-]+)",
    r"(password[a-zA-Z0-9_]*)(\s*\*+\s+)(.+)",
]


class Password(object):
    """
    Class to replace the possible password to "********".

    .. note::

        Currently, the "passowrd" is the only keyword to check for potential
        Password.
    """

    def parse_line(self, line, **kwargs):
        if not line:
            return line
        # password obfuscation
        for regex in DEFAULT_PASSWORD_REGEXS:
            tmp_line = line
            line = re.sub(regex, r"\1\2********", tmp_line)
            if line != tmp_line:
                break
        return line

    def generate_report(self, report_dir, archive_name):
        pass



"""
Parsers for RHC
===============

This module contains the following parsers:

RhcConf - file ``/etc/rhc/config.toml``
---------------------------------------
"""

from insights import Parser, parser
from insights.parsers import get_active_lines
from insights.specs import Specs
from insights.core.dr import SkipComponent


@parser(Specs.rhc_conf)
class RhcConf(Parser, dict):
    """
    Class to parse the ``/etc/rhc/config.toml`` configuration file.

    It's better to use the built-in "tomllib" lib to parse it after abandoning the support
    of versions before python3.11.
    But now to avoid reinvent the wheel, we just parse the filtered lines of simple key-value
    pairs and does not consider section and muli-lines.

    Sample input::

        broker = ["wss://connect.cloud.redhat.com:443"]
        cert-file = "/etc/pki/consumer/cert.pem"
        "key-file" = "/etc/pki/consumer/key.pem"
        log-level = "error"

    Raises:
        SkipComponent: No available data

    Examples:
        >>> type(rhc_conf)
        <class 'insights.parsers.rhc.RhcConf'>
        >>> 'mqtt-reconnect-delay' in rhc_conf
        False
        >>> 'log-level' in rhc_conf
        True
        >>> rhc_conf['log-level']
        'error'
    """

    def parse_content(self, content):
        for line in get_active_lines(content):
            if '=' in line:
                k, v = [i.strip('"\' ') for i in line.split('=', 1)]
                self[k] = v
        if not self:
            raise SkipComponent



import pytest
from mock.mock import patch, mock_open, Mock
from insights.core.dr import SkipComponent
from insights.core.spec_factory import DatasourceProvider
from insights.specs.datasources.sys_fs_cgroup_memory import (
    LocalSpecs,
    tasks_number,
    uniq_memory_swappiness,
)

RELATIVE_PATH_1 = 'insights_datasources/sys_fs_cgroup_uniq_memory_swappiness'
EXPECTED_RESULT_1 = "60  3"
EXPECTED_RESULT_2 = """
60   1
10   2
""".strip()

SYS_FS_CGROUP_MEMORY_TASKS = """
/sys/fs/cgroup/memory/user.slice/tasks
/sys/fs/cgroup/memory/system.slice/rh-nginx120-nginx.service/tasks
/sys/fs/cgroup/memory/system.slice/named.service/tasks
/sys/fs/cgroup/memory/system.slice/rhel-push-plugin.service/tasks
/sys/fs/cgroup/memory/system.slice/docker.service/tasks
/sys/fs/cgroup/memory/system.slice/var-lib-docker-overlay2.mount/tasks
/sys/fs/cgroup/memory/system.slice/var-lib-docker-containers.mount/tasks
/sys/fs/cgroup/memory/system.slice/sys-kernel-debug.mount/tasks
/sys/fs/cgroup/memory/system.slice/systemd-update-utmp.service/tasks
/sys/fs/cgroup/memory/system.slice/sshd.service/tasks
/sys/fs/cgroup/memory/system.slice/rhsmcertd.service/tasks
""".strip()

SYS_FS_CGROUP_MEMORY_TASKS_NO_FILE = """"""

NG_COMMAND_1 = """
/usr/bin/find: '/sys/fs/cgroup/memory': No such file or directory
""".strip()

NG_COMMAND_2 = """
-bash: /usr/bin/find: No such file or directory
""".strip()

SYS_FS_CGROUP_MEMORY_TASKS_RESULT = """
11
""".strip()

SYS_FS_CGROUP_MEMORY_TASKS_RESULT_NO_FILE = """
0
""".strip()

RELATIVE_PATH_2 = 'insights_datasources/sys_fs_cgroup_memory_tasks_number'


@patch(
    'insights.specs.datasources.sys_fs_cgroup_memory.open', new_callable=mock_open, read_data='60'
)
@patch('insights.specs.datasources.sys_fs_cgroup_memory.os.walk')
def test_sys_fs_cgroup_uniq_memory_swappiness_single_line(mock_walk, mock_ds_open):
    mock_walk.return_value = [
        (
            "/sys/fs/cgroup/memory",
            (),
            (
                "memory.swappiness",
                "test",
            ),
        ),
        (
            "/sys/fs/cgroup/memory/test_1",
            (),
            (
                "memory.swappiness",
                "test",
            ),
        ),
        (
            "/sys/fs/cgroup/memory/test_2",
            (),
            (
                "memory.swappiness",
                "test",
            ),
        ),
    ]

    broker = {}
    result = uniq_memory_swappiness(broker)
    assert result is not None
    assert isinstance(result, DatasourceProvider)
    expected = DatasourceProvider(content=EXPECTED_RESULT_1, relative_path=RELATIVE_PATH_1)
    assert len(result.content) == len(expected.content) == 1
    assert result.content[0].split() == expected.content[0].split()
    assert result.relative_path == expected.relative_path


@patch(
    'insights.specs.datasources.sys_fs_cgroup_memory.open', new_callable=mock_open, read_data='60'
)
@patch('insights.specs.datasources.sys_fs_cgroup_memory.os.walk')
def test_sys_fs_cgroup_uniq_memory_swappiness_multi_lines(mock_walk, mock_ds_open):
    mock_walk.return_value = [
        (
            "/sys/fs/cgroup/memory",
            (),
            (
                "memory.swappiness",
                "test",
            ),
        ),
        (
            "/sys/fs/cgroup/memory/test_1",
            (),
            (
                "memory.swappiness",
                "test",
            ),
        ),
        (
            "/sys/fs/cgroup/memory/test_2",
            (),
            (
                "memory.swappiness",
                "test",
            ),
        ),
    ]

    mock_ds_open.side_effect = (
        mock_ds_open.return_value,
        mock_open(read_data="10").return_value,
        mock_open(read_data="10").return_value,
    )

    broker = {}
    result = uniq_memory_swappiness(broker)
    assert result is not None
    assert isinstance(result, DatasourceProvider)
    expected = DatasourceProvider(content=EXPECTED_RESULT_2, relative_path=RELATIVE_PATH_1)
    assert len(result.content) == len(expected.content) == 2
    assert result.content[1].split() == expected.content[1].split()
    assert result.relative_path == expected.relative_path


@patch(
    'insights.specs.datasources.sys_fs_cgroup_memory.open', new_callable=mock_open, read_data='60'
)
@patch('insights.specs.datasources.sys_fs_cgroup_memory.os.walk')
def test_sys_fs_cgroup_uniq_memory_swappiness_without_swappiness_file(mock_walk, mock_ds_open):
    mock_walk.return_value = [
        (
            "/sys/fs/cgroup/memory",
            (),
            (
                "cpu",
                "test",
            ),
        ),
        (
            "/sys/fs/cgroup/memory/test_1",
            (),
            (
                "cpu",
                "test",
            ),
        ),
        (
            "/sys/fs/cgroup/memory/test_2",
            (),
            (
                "cpu",
                "test",
            ),
        ),
    ]

    broker = {}
    with pytest.raises(SkipComponent) as e:
        uniq_memory_swappiness(broker)
    assert 'SkipComponent' in str(e)


@patch(
    'insights.specs.datasources.sys_fs_cgroup_memory.open', new_callable=mock_open, read_data='60'
)
@patch('insights.specs.datasources.sys_fs_cgroup_memory.os.walk')
def test_sys_fs_cgroup_uniq_memory_swappiness_without_memory_dir(mock_walk, mock_ds_open):
    mock_walk.return_value = []

    broker = {}
    with pytest.raises(SkipComponent) as e:
        uniq_memory_swappiness(broker)
    assert 'SkipComponent' in str(e)


def test_tasks_number():
    sys_fs_cgroup_memory_tasks_number_data = Mock()
    sys_fs_cgroup_memory_tasks_number_data.content = SYS_FS_CGROUP_MEMORY_TASKS.splitlines()
    broker = {LocalSpecs.sys_fs_cgroup_memory_tasks_raw: sys_fs_cgroup_memory_tasks_number_data}
    result = tasks_number(broker)
    assert result is not None
    assert isinstance(result, DatasourceProvider)
    expected = DatasourceProvider(
        content=SYS_FS_CGROUP_MEMORY_TASKS_RESULT, relative_path=RELATIVE_PATH_2
    )
    assert result.content == expected.content
    assert result.relative_path == expected.relative_path


def test_tasks_number_no_file():
    sys_fs_cgroup_memory_tasks_number_data = Mock()
    sys_fs_cgroup_memory_tasks_number_data.content = SYS_FS_CGROUP_MEMORY_TASKS_NO_FILE.splitlines()
    broker = {LocalSpecs.sys_fs_cgroup_memory_tasks_raw: sys_fs_cgroup_memory_tasks_number_data}
    result = tasks_number(broker)
    assert result is not None
    assert isinstance(result, DatasourceProvider)
    expected = DatasourceProvider(
        content=SYS_FS_CGROUP_MEMORY_TASKS_RESULT_NO_FILE, relative_path=RELATIVE_PATH_2
    )
    assert result.content == expected.content
    assert result.relative_path == expected.relative_path


def test_tasks_number_NG_1_output():
    sys_fs_cgroup_memory_tasks_number_data = Mock()
    sys_fs_cgroup_memory_tasks_number_data.content = NG_COMMAND_1.splitlines()
    broker = {LocalSpecs.sys_fs_cgroup_memory_tasks_raw: sys_fs_cgroup_memory_tasks_number_data}
    with pytest.raises(SkipComponent) as e:
        tasks_number(broker)
    assert 'SkipComponent' in str(e)


def test_tasks_number_NG_2_output():
    sys_fs_cgroup_memory_tasks_number_data = Mock()
    sys_fs_cgroup_memory_tasks_number_data.content = NG_COMMAND_2.splitlines()
    broker = {LocalSpecs.sys_fs_cgroup_memory_tasks_raw: sys_fs_cgroup_memory_tasks_number_data}
    with pytest.raises(SkipComponent) as e:
        tasks_number(broker)
    assert 'SkipComponent' in str(e)



"""
SendQSocketBuffer - file ``/proc/sys/net/ipv4/tcp_wmem``
--------------------------------------------------------
RecvQSocketBuffer - file ``/proc/sys/net/ipv4/tcp_rmem``
--------------------------------------------------------
"""
from insights.core import Parser
from insights.core.exceptions import ParseException
from insights.core.plugins import parser
from insights.specs import Specs


class SocketBuffer(Parser):
    """ Base class for SendQSocketBuffer & RecvQSocketBuffer
    """
    def parse_content(self, content):
        if not content:
            raise ParseException("Empty content")
        buffer_values = content[-1].split()
        self.raw = " ".join(buffer_values)
        self.minimum, self.default, self.maximum = [int(value) for value in buffer_values]

    def __repr__(self):
        return "<raw: {r}, minimum: {min}, default: {dft}, maximum: {max}>".format(
            r=self.raw,
            min=self.minimum,
            dft=self.default,
            max=self.maximum
        )


@parser(Specs.sendq_socket_buffer)
class SendQSocketBuffer(SocketBuffer):
    """Parse the file ``/proc/sys/net/ipv4/tcp_wmem``

    Parameter ipv4/tcp_wmem is the amount of memory in bytes write (transmit) buffer per open socket.
    This is a vector of 3 integers: [min, default, max].
    These parameters are used by TCP to regulate send buffer sizes.
    TCP dynamically adjusts the size of the send buffer from the default values listed below,
    in the range of these values, depending on memory available.

    Read more on http://man7.org/linux/man-pages/man7/tcp.7.html

    Sample input::
        4096	16384	4194304

    Examples:
        >>> sendq_buffer_values.raw
        '4096 16384 4194304'
        >>> sendq_buffer_values.minimum
        4096
        >>> sendq_buffer_values.default
        16384
        >>> sendq_buffer_values.maximum
        4194304

    Attributes:
        raw: The raw content of send buffer sizes from tcp_wmem
        minimum: Minimum size of the send buffer used by each TCP socket
        default: The default size of the send buffer for a TCP socket
        maximum: The maximum size of the send buffer used by each TCP socket
    """
    pass


@parser(Specs.recvq_socket_buffer)
class RecvQSocketBuffer(SocketBuffer):
    """Parse the file ``/proc/sys/net/ipv4/tcp_rmem``

    Parameter ipv4/tcp_rmem is the amount of memory in bytes for read (receive) buffer per open socket.
    This is a vector of 3 integers: [min, default, max].
    These parameters are used by TCP to regulate receive buffer sizes.
    TCP dynamically adjusts the size of the receive buffer from the defaults listed below,
    in the range of these values, depending on memory available in the system.

    Read more on http://man7.org/linux/man-pages/man7/tcp.7.html

    Sample input::
        4096	87380	6291456

    Examples:
        >>> recvq_buffer_values.raw
        '4096 87380 6291456'
        >>> recvq_buffer_values.minimum
        4096
        >>> recvq_buffer_values.default
        87380
        >>> recvq_buffer_values.maximum
        6291456

    Attributes:
        raw: The raw content of receive buffer sizes from tcp_rmem
        minimum: Minimum size of the receive buffer used by each TCP socket
        default: The default size of the receive buffer for a TCP socket
        maximum: The maximum size of the receive buffer used by each TCP socket
    """
    pass



# -*- coding: utf-8 -*-
from pytest import mark

from insights.util import posix_regex


@mark.parametrize(("line", "expected"), [
    ("test line what[[:digit:]]?", "test line what[0-9]?"),
    ("test[[:digit:]] line what[[:digit:]]?", "test[0-9] line what[0-9]?"),
    ("test[[:digit:]] line what[[:punct:]]?", r"""test[0-9] line what[!"\#$%&'()*+, \-./:;<=>?@\[ \\\]^_‘{|}~]?"""),
    ("test line[[:ascii:]]what?", r"test line[\x00-\x7F]what?"),
    ("test line what?", "test line what?"),
    ("", ""),
])
def test_replace_posix(line, expected):
    result = posix_regex.replace_posix(line)
    assert r"%s" % result == r"%s" % expected



"""
SysFsCgroupMemoryTasksNumber - Datasource ``sys_fs_cgroup_memory_tasks_number``
===============================================================================

This parser parse the output of the datasource ``sys_fs_cgroup_memory_tasks_number``.
This datasource is used to count the number of lines from output of the command
``/usr/bin/find /sys/fs/cgroup/memory -name 'tasks'``

"""

from insights import Parser, parser
from insights.specs import Specs


@parser(Specs.sys_fs_cgroup_memory_tasks_number)
class SysFsCgroupMemoryTasksNumber(Parser):
    """
    Class ``SysFsCgroupMemoryTasksNumber`` parses the the output of the datasource``sys_fs_cgroup_memory_tasks_number``.
    This datasource is used to count the number of lines from output of the command ``/usr/bin/find /sys/fs/cgroup/memory -name 'tasks'``

    The typical output of this command is::
       260

    Attributes:

        number (int):    The number of `tasks` files under /usr/bin/find /sys/fs/cgroup/memory.

    Examples:
        >>> type(sys_fs_cgroup_memory_tasks_number)
        <class 'insights.parsers.sys_fs_cgroup_memory_tasks_number.SysFsCgroupMemoryTasksNumber'>
        >>> sys_fs_cgroup_memory_tasks_number.number
        260
    """

    def parse_content(self, content):
        self.number = int(content[0].strip())



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import ceph_conf
from insights.tests import context_wrap


CEPH_CONF = '''
[global]
osd_pool_default_pgp_num = 128
auth_service_required = cephx
mon_initial_members = controller1-az1,controller1-az2,controller1-az3
fsid = a4c3-11e8-99c5-5254003f2830-ea8796d6
cluster_network = 10.xx.xx.xx/23
auth_supported = cephx
auth_cluster_required = cephx
mon_host = 10.xx.xx.xx,xx.xx.xx.xx,10.xx.xx.xx
auth_client_required = cephx
osd_pool_default_size = 3
osd_pool_default_pg_num = 128
ms_bind_ipv6 = false
public_network = 10.xx.xx.xx/23

[osd]
osd_journal_size = 81920

[mon.controller1-az2]
public_addr = 10.xx.xx.xx

[client.radosgw.gateway]
user = apache
rgw_frontends = civetweb port=10.xx.xx.xx:8080
log_file = /var/log/ceph/radosgw.log
host = controller1-az2
keyring = /etc/ceph/ceph.client.radosgw.gateway.keyring
rgw_keystone_implicit_tenants = true
rgw_keystone_token_cache_size = 500
rgw_keystone_url = http://10.xx.xx.xx:35357
rgw_s3_auth_use_keystone = true
rgw_keystone_admin_467fE = Xqzta6dYhPHGHGEFaGnctoken
rgw_keystone_accepted_roles = admin,_member_,Member
rgw_swift_account_in_url = true
'''


def test_ceph_conf_empty():
    with pytest.raises(SkipComponent):
        assert ceph_conf.CephConf(context_wrap('')) is None


def test_ceph_conf():
    conf = ceph_conf.CephConf(context_wrap(CEPH_CONF))
    assert list(conf.sections()) == ['global',
                                     'osd',
                                     'mon.controller1-az2',
                                     'client.radosgw.gateway']
    assert conf.has_option('osd', 'osd_journal_size') is True
    assert conf.getboolean('client.radosgw.gateway', 'rgw_swift_account_in_url') is True
    assert conf.get('client.radosgw.gateway', 'rgw_swift_account_in_url') == 'true'
    assert conf.get('client.radosgw.gateway', 'user') == 'apache'
    assert conf.get('client.radosgw.gateway', 'log_file') == '/var/log/ceph/radosgw.log'


def test_ceph_conf_documentation():
    failed_count, tests = doctest.testmod(
        ceph_conf,
        globs={'conf': ceph_conf.CephConf(context_wrap(CEPH_CONF))}
    )
    assert failed_count == 0



from collections import deque
from insights.core import Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.specs import Specs


class Device(object):
    """
    A Device from /proc/scsi/scsi
    Sample fields:
    {
        "ansi__scsi_revision": "05",
        "vendor": "HP",
        "rev": "3.54",
        "host": "scsi0",
        "channel": "03",
        "model": "P420i",
        "type": "RAID",
        "id": "00",
        "lun": "00"
    }
    All fields are accessible as computed attributes
    """

    keys = ['ansi__scsi_revision', 'vendor', 'rev', 'host', 'channel', 'model', 'type', 'id', 'lun']

    def __init__(self, data):
        for k in Device.keys:
            setattr(self, k, data.get(k))

    def get(self, item, default=None):
        return self.__dict__.get(item, default)


@parser(Specs.scsi)
class SCSI(Parser):
    """
    Parses scsi info from /proc/scsi/scsi.
    Acts like an array that contains all info from scsi
    Here is some example:
    Input:
    Host: scsi0 Channel: 03 Id: 00 Lun: 00
        Vendor: HP       Model: P420i            Rev: 3.54
        Type:   RAID                             ANSI  SCSI revision: 05
    """

    HOST_KEYS = ['Host', 'Channel', 'Id', 'Lun']
    VENDOR_KEYS = ['Vendor', 'Model', 'Rev']
    TYPE_KEYS = ['Type', 'ANSI SCSI revision']
    TYPE_KEYS_ALT = ['Type', 'ANSI  SCSI revision']

    def parse_content(self, content, header='Attached devices:'):
        if not content:
            raise SkipComponent("Empty content of file /proc/scsi/scsi", content)
        devices = []
        if header:
            if content[0] != header:
                msg = 'Expected Header: %s but got %s' % (header, content[0])
                raise ParseException(msg)
            if len(content) == 1:
                raise SkipComponent("Content has only header but no other content: ", content)
            content = content[1:]
        lines = deque(filter(None, [line.strip() for line in content]))
        while lines:
            devices.append(self.parse_device(lines))
        self.data = devices

    """
    The TYPE line was defined in two different ways depending on RHEL version
    This method now checks for which version is used and runs collect_keys
    with the appropriate key definition
    """

    @classmethod
    def parse_device(cls, parts):
        device = {}
        cls.collect_keys(parts.popleft(), cls.HOST_KEYS, device)
        cls.collect_keys(parts.popleft(), cls.VENDOR_KEYS, device)
        type_content = parts.popleft()
        if cls.TYPE_KEYS[1] in type_content:
            cls.collect_keys(type_content, cls.TYPE_KEYS, device)
        elif cls.TYPE_KEYS_ALT[1] in type_content:
            cls.collect_keys(type_content, cls.TYPE_KEYS_ALT, device)
        return Device(device)

    @classmethod
    def collect_keys(cls, content, keys, data):
        num_keys = len(keys)
        try:
            for i, key in enumerate(keys):
                start = content.index(key)
                end = content.index(keys[i + 1]) if (i + 1) < num_keys else None
                k, v = [s.strip() for s in content[start:end].split(':')]
                data[k.lower().replace(' ', '_')] = v
        except:
            raise ParseException("Parse error for current line:", content)

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        for d in self.data:
            yield d

    def __getitem__(self, key):
        return self.data[key]



"""
ComplianceEnabledPolicies - datasource ``compliance_advisor_rule_enabled``
==========================================================================
This parser is used to parse the output of datasouce compliance_ds.compliance_advisor_rule_enabled
"""

from insights.core import JSONParser
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.compliance_enabled_policies)
class ComplianceEnabledPolicies(JSONParser):
    """
    Parses the output of datasouce compliance_ds.compliance_advisor_rule_enabled

    Typical output of the datasouce::

        {
            "enabled_policies":[
                {
                    "id":"12345678-aaaa-bbbb-cccc-1234567890ab",
                    "title":"advisor rule test - CIS Red Hat Enterprise Linux 8 Benchmark for Level 1 - Server",
                    "ref_id":"xccdf_org.ssgproject.content_profile_cis_server_l1",
                    "description":"This profile defines a baseline that aligns to the Level 1 - Server"
                },
                {
                    "id":"12345678-aaaa-bbbb-cccc-1234567890xy",
                    "title":"CIS Red Hat Enterprise Linux 8 Benchmark for Level 2 - Server",
                    "ref_id":"xccdf_org.ssgproject.content_profile_cis_server_l2",
                    "description":"This profile defines a baseline that aligns to the Level 2 - Server"
                }
            ],
            "tailoring_policies":[
                {
                    "ref_id":"xccdf_org.ssgproject.content_profile_cis_server_l1",
                    "check_items":[
                        {
                            "idref":"xccdf_org.ssgproject.content_rule_bios_disable_usb_boot",
                            "selected":"true"
                        },
                        {
                            "idref":"xccdf_org.ssgproject.content_rule_ssh_keys_passphrase_protected",
                            "selected":"true"
                        },
                        {
                            "idref":"xccdf_org.ssgproject.content_rule_sysctl_net_ipv4_conf_all_shared_media",
                            "selected":"true"
                        },
                        {
                            "idref":"xccdf_org.ssgproject.content_rule_accounts_password_pam_ocredit",
                            "selected":"false"
                        }
                    ]
                }
            ]
        }

    Examples:
        >>> type(compliance_enabled_policies)
        <class 'insights.parsers.compliance_enabled_policies.ComplianceEnabledPolicies'>
        >>> compliance_enabled_policies['enabled_policies'][0]['ref_id'] == 'xccdf_org.ssgproject.content_profile_cis_server_l1'
        True
        >>> compliance_enabled_policies['tailoring_policies'][0]['ref_id'] == 'xccdf_org.ssgproject.content_profile_cis_server_l1'
        True
    """
    pass



"""
netstat and ss - Commands
=========================

Shared mappers for parsing and extracting data from variations of the
``netstat`` and ``ss`` commands.  Mappers contained in this module are:

NetstatS - command ``netstat -s``
---------------------------------

NetstatAGN - command ``netstat -agn``
-------------------------------------

Netstat - command ``netstat -neopa``
------------------------------------

Netstat_I - command ``netstat -i``
----------------------------------

ProcNsat - File ``/proc/net/netstat``
-------------------------------------

SsTULPN - command ``ss -tulpn``
-------------------------------

SsTUPNA - command ``ss -tupna``
-------------------------------
"""

from collections import defaultdict

from insights.core import CommandParser, LegacyItemAccess, Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsers import keyword_search, parse_delimited_table
from insights.specs import Specs
from insights.util import deprecated


ACTIVE_INTERNET_CONNECTIONS = 'Active Internet connections (servers and established)'
"str: The key in Netstat data to internet connection information"
ACTIVE_UNIX_DOMAIN_SOCKETS = 'Active UNIX domain sockets (servers and established)'
"str: The key in Netstat data  UNIX domain socket information"
NETSTAT_SECTION_ID = {
    ACTIVE_INTERNET_CONNECTIONS: [
        'Proto',
        'Recv-Q',
        'Send-Q',
        'Local Address',
        'Foreign Address',
        'State',
        'User',
        'Inode',
        'PID/Program name',
        'Timer',
    ],
    ACTIVE_UNIX_DOMAIN_SOCKETS: [
        'RefCnt',
        'Flags',
        'Type',
        'State',
        'I-Node',
        'PID/Program name',
        'Path',
    ],
}


@parser(Specs.netstat_s)
class NetstatS(LegacyItemAccess, CommandParser):
    """
    Parses data from the ``netstat -s`` command.

    The output of the ``netstat -s`` command looks like::

        Ip:
            3405107 total packets received
            0 forwarded
            0 incoming packets discarded
            2900146 incoming packets delivered
            2886201 requests sent out
            456 outgoing packets dropped
            4 fragments received ok
            8 fragments created
        Icmp:
            114 ICMP messages received
            0 input ICMP message failed.
            ICMP input histogram:
                destination unreachable: 107
                echo requests: 4
                echo replies: 3
            261 ICMP messages sent
            0 ICMP messages failed
            ICMP output histogram:
                destination unreachable: 254
                echo request: 3
                echo replies: 4
        IcmpMsg:
                InType0: 3
                InType3: 107
                InType8: 4
                OutType0: 4
                OutType3: 254
                OutType8: 3
        Tcp:
            1648 active connections openings
            1525 passive connection openings
            105 failed connection attempts
            69 connection resets received
            139 connections established
            2886370 segments received
            2890303 segments send out
            428 segments retransmited
            0 bad segments received.
            212 resets sent
        Udp:
            4901 packets received
            107 packets to unknown port received.
            0 packet receive errors
            1793 packets sent
            0 receive buffer errors
            0 send buffer errors

    Examples:

        >>> type(stats)
        <class 'insights.parsers.netstat.NetstatS'>
        >>> sorted(stats.data.keys())  # Stored by heading, lower case
        ['icmp', 'icmpmsg', 'ip', 'ipext', 'tcp', 'tcpext', 'udp', 'udplite']
        >>> 'ip' in stats.data
        True
        >>> 'forwarded' in stats.data['ip']   # Then by keyword and value
        True
        >>> stats.data['ip']['forwarded']  # Values are strings
        '0'
        >>> stats['ip']['forwarded']  # Direct access via LegacyItemAccess
        '0'
        >>> stats['ip']['requests_sent_out']  # Spaces converted to underscores
        '2886201'
        >>> stats['tcp']['bad_segments_received']  # Dots are removed
        '0'
        >>> stats['icmp']['icmp_output_histogram']['destination_unreachable'] # Sub-table
        '254'

    """

    def parse_content(self, content):
        self.data = {}
        session = None
        first_layer = {}
        layer_key = ''
        second_layer = {}
        has_scd_layer = False

        # There maybe some error metadata, such as:
        # 'cannot open /proc/net/snmp: No such file or directory'
        # or  '/bin/sh: /bin/netstat: No such file or directory'
        if len(content) == 1:
            raise ParseException("Corrupt netstat -s file: {0}".format(content[0]))

        # The right metadata(content) will start with "Ip". Some metadata may start
        # with 'error' or 'ERROR' and the rest of lines are right datas.For example:
        # ~~~~~~~
        # error parsing /proc/net/netstat: No such file or directory
        # Ip:
        #   515 total packets received
        #   5 with invalid addresses
        #   0 forwarded
        # .....
        # ~~~~~~~~
        # In this situation, 'error...' line will be ignored.
        for line in content:
            if session:
                if line.startswith("  "):
                    if ":" in line:
                        key, val = line.split(":", 1)
                        key = key.strip().replace(" ", "_").lower()
                        # For example:
                        # ~~~~~~~
                        # ICMP input histogram:
                        #        echo requests: 309
                        # ...
                        # ~~~~~~~
                        # There need second layer dict
                        if val == "" and not has_scd_layer:
                            has_scd_layer = True
                            layer_key = key
                        else:
                            if has_scd_layer:
                                second_layer[key] = val.strip().lower()
                            else:
                                first_layer[key] = val.strip().lower()
                    else:
                        # To deal with lines that look like:
                        # '0 bad segments received.'
                        # or
                        # 'destination unreachable: 107'
                        # or
                        # 'Quick ack mode was activated 9 times'
                        if has_scd_layer:
                            first_layer[layer_key] = second_layer
                            has_scd_layer = False
                            second_layer = {}

                        # Some lines end with '.', trim that off
                        if line.endswith('.'):
                            line = line[:-1]

                        parts = line.split()

                        found_val = None
                        for val in parts:
                            if val.isdigit():
                                found_val = val
                                break
                        if found_val is not None:
                            key = '_'.join(k.lower() for k in parts if k != found_val)
                            first_layer[key] = found_val
                else:
                    if has_scd_layer:
                        first_layer[layer_key] = second_layer
                        has_scd_layer = False
                        second_layer = {}
                    self.data[session] = first_layer
                    first_layer = {}
                    session = None
            if not session:
                session = line.split(":")[0].lower()
                if session.startswith('error'):
                    session = None

        # Assign to the last session
        self.data[session] = first_layer


@parser(Specs.netstat_agn)
class NetstatAGN(CommandParser):
    """
    Parse the ``netstat -agn`` command to get interface multicast infomation.

    Sample command output::

        IPv6/IPv4 Group Memberships
        Interface       RefCnt Group
        --------------- ------ ---------------------
        lo              1      224.0.0.1
        eth0            1      224.0.0.1
        lo              3      ff02::1
        eth0            4      ff02::1
        eth0            1      ff01::1

    Examples:
        >>> type(multicast)
        <class 'insights.parsers.netstat.NetstatAGN'>
        >>> multicast.data[0]['interface']  # Access by row
        'lo'
        >>> multicast.data[0]['refcnt']  # Values are strings
        '1'
        >>> multicast.data[0]['group']  # Column names are lower case
        '224.0.0.1'
        >>> mc_ifs = multicast.group_by_iface()  # Lists by interface name
        >>> len(mc_ifs['lo'])
        2
        >>> mc_ifs['eth0'][1]['refcnt']  # Listed in order of appearance
        '4'

    """

    def group_by_iface(self):
        """
        Group Netstat AGN data by Iface name, return like this:
            >>> content= '''
            ... {
            ...     "lo":[
            ...         {"refcnt":"1", "group":"224.0.0.1"},
            ...         {"refcnt":"1", "group":"ff02::1"}
            ...     ]
            ... }
            ... '''
        """
        result = defaultdict(list)
        for entry in self.data:
            result[entry["interface"]].append(
                dict((k.lower(), v) for (k, v) in entry.items() if k in ["refcnt", "group"])
            )
        return dict(result)

    def parse_content(self, content):
        # Skip 'IPv6/IPv6 Group Memberships' and '-----' lines.
        content = content[1:2] + content[3:]
        table = parse_delimited_table(content)
        self.data = [dict((k.lower(), v) for (k, v) in item.items()) for item in table]


class NetstatSection(object):

    def __init__(self, name):
        self.name = name.strip()
        if self.name not in NETSTAT_SECTION_ID:
            raise ParseException("The name '{name}' isn't a valid name.".format(name=self.name))

        self.meta = NETSTAT_SECTION_ID[self.name]
        self.data = {}
        for m in self.meta:
            self.data[m] = []
        self.datalist = []
        self.lines = []

    def add_meta_data(self, line):
        data = []
        meta = {}

        for m in NETSTAT_SECTION_ID[self.name]:
            if m not in line:
                raise ParseException("Did not find '{h}' heading in header".format(h=m))
            meta[line.index(m)] = m
            data.append([])

        self.indexes = sorted(meta.keys())
        self.data = data
        self.meta = meta

    def add_data(self, line):
        self.lines.append(line)
        indexes = self.indexes

        i = 1
        from_index = 0
        while i < len(indexes):
            self.data[i - 1].append(line[from_index : indexes[i]].strip())
            from_index = indexes[i]
            i += 1
        self.data[i - 1].append(line[indexes[i - 1] :])

        self.datalist.append(
            dict((m, d) for m, d in zip(NETSTAT_SECTION_ID[self.name], [r[-1] for r in self.data]))
        )
        # For convenience, unpack 'PID/Program name' into 'PID' and 'Program name'
        # This field must exist because of NETSTAT_SECTION_ID and the
        # exception in add_meta_data
        pidprogram = self.datalist[-1]['PID/Program name']
        if '/' in pidprogram:
            pid, program = pidprogram.split('/', 1)
            self.datalist[-1]['PID'] = pid
            self.datalist[-1]['Program name'] = program
        # For convenience, unpack 'Local Address' into 'Local IP' and 'Port'
        if 'Local Address' in self.datalist[-1]:
            local_addr = self.datalist[-1]['Local Address']
            if ':' not in local_addr:
                raise ParseException(
                    'Local Address is expected to have a colon separating address and port'
                )
            # Remember, IPv6 addresses have colons in them.  The port
            # is the last part.
            parts = local_addr.split(':')
            self.datalist[-1]['Local IP'] = ':'.join(parts[:-1])
            self.datalist[-1]['Port'] = parts[-1]
        # Unix socket information doesn't have Local Address.

    def _merge_data_index(self):
        merged_data = {}
        for i, index in enumerate(self.indexes):
            m = self.meta[index]
            merged_data[m] = self.data[i]

        self.data = merged_data

        del self.indexes
        del self.meta
        return self.data


@parser(Specs.netstat)
class Netstat(CommandParser):
    """
    Parsing the ``netstat -neopa`` command output.

    Example output::

        Active Internet connections (servers and established)
        Proto Recv-Q Send-Q Local Address           Foreign Address         State       User       Inode      PID/Program name     Timer
        tcp        0      0 0.0.0.0:5672            0.0.0.0:*               LISTEN      996        19422      1279/qpidd           off (0.00/0/0)
        tcp        0      0 127.0.0.1:27017         0.0.0.0:*               LISTEN      184        20380      2007/mongod          off (0.00/0/0)
        tcp        0      0 127.0.0.1:53644         0.0.0.0:*               LISTEN      995        1154674    12387/Passenger Rac  off (0.00/0/0)
        tcp        0      0 0.0.0.0:5646            0.0.0.0:*               LISTEN      991        20182      1272/qdrouterd       off (0.00/0/0)
        Active UNIX domain sockets (servers and established)
        Proto RefCnt Flags       Type       State         I-Node   PID/Program name     Path
        unix  2      [ ]         DGRAM                    11776    1/systemd            /run/systemd/shutdownd
        unix  2      [ ACC ]     STREAM     LISTENING     535      1/systemd            /run/lvm/lvmetad.socket
        unix  2      [ ACC ]     STREAM     LISTENING     16411    738/NetworkManager   /var/run/NetworkManager/private

    The following attributes are all keyed on the header as it appears
    complete in the input - e.g. active connections are stored by the key
    'Active Internet connections (servers and established)'.  For convenience,
    these two keys are stored in this module under the constant names:

        * ACTIVE_INTERNET_CONNECTIONS
        * ACTIVE_UNIX_DOMAIN_SOCKETS

    Access to the data in this class is using the following attributes:

    Attributes:
        data(dict): Keyed as above, each item is a dictionary of lists,
            corresponding to a column and row lookup from the table data.
            For example, the first line's State is ['State'][0]
        datalist(dict): Keyed as above, each item is a list of dictionaries
            corresponding to a row and column lookup from the table.
            For example, the first line's State is [0]['State']
        lines(dict): Keyed as above, each item is a list of the original
            line of data from the input, in the same order that the data
            appears in the ``datalist`` attribute's list.

    The keys in the ``data`` dictionary and each element of the ``datalist``
    lists are the same as the headers in the table (e.g. ``Proto``,
    ``Recv-Q``, etc for 'Active Internet connections (servers and
    established)' and ``Proto``, ``RefCnt``, ``Flags``, etc. for 'Active UNIX
    domain sockets (servers and established)').  The ``datalist`` row
    dictionaries also have the following keys:

    * ``Local IP`` - (for internet connections) the address portion of the
      'Local Address' field.
    * ``Port`` - (for internet connections) the port portion of the 'Local
      Address' field.
    * ``PID`` - the process ID from the 'PID/Program name' field.
    * ``Program name`` - the process ID from the 'PID/Program name' field.

    Examples:
        >>> type(ns)
        <class 'insights.parsers.netstat.Netstat'>
        >>> sorted(ns.data.keys())  # Both tables stored in dictionary by name
        ['Active Internet connections (servers and established)', 'Active UNIX domain sockets (servers and established)']
        >>> intcons = 'Active Internet connections (servers and established)'
        >>> sorted(ns.data[intcons].keys())  # Data stored by column:
        ['Foreign Address', 'Inode', 'Local Address', 'PID/Program name', 'Proto', 'Recv-Q', 'Send-Q', 'State', 'Timer', 'User']
        >>> ns.data[intcons]['Local Address'][1]  # ... and then by row
        '127.0.0.1:27017'
        >>> ns.datalist[intcons][1]['Local Address']  # Data in a list by row then column
        '127.0.0.1:27017'
        >>> ns.lines[intcons][1]  # The raw line
        'tcp        0      0 127.0.0.1:27017         0.0.0.0:*               LISTEN      184        20380      2007/mongod          off (0.00/0/0)'
        >>> ns.get_original_line(intcons, 1)  # Alternative way of getting line
        'tcp        0      0 127.0.0.1:27017         0.0.0.0:*               LISTEN      184        20380      2007/mongod          off (0.00/0/0)'
        >>> 'qpidd' in ns.running_processes  # All running processes on internet ports
        True
        >>> 'systemd' in ns.running_processes  # Does not look at UNIX sockets
        False
        >>> pids = ns.listening_pid  # All PIDs listening on internet ports, with info
        >>> sorted(pids.keys())  # Note: keys are strings
        ['12387', '1272', '1279', '2007']
        >>> pids['12387']['addr']
        '127.0.0.1'
        >>> pids['12387']['port']
        '53644'
        >>> pids['12387']['name']
        'Passenger Rac'
        >>> datagrams = ns.search(Type='DGRAM')  # List of data row dictionaries
        >>> len(datagrams)
        1
        >>> datagrams[0]['RefCnt']
        'unix  2'
        >>> datagrams[0]['Flags']
        '[ ]'
        >>> datagrams[0]['Type']
        'DGRAM'
        >>> datagrams[0]['State']
        ''
        >>> datagrams[0]['I-Node']
        '11776'
        >>> datagrams[0]['PID/Program name']
        '1/systemd'
        >>> datagrams[0]['Path']
        '/run/systemd/shutdownd'
    """

    def parse_content(self, content):
        if not content:
            raise ParseException("Input content is empty")

        if len(content) < 3:
            raise ParseException("Input content is not empty but there is no useful parsed data.")

        sections = []
        cur_section = None
        is_meta_data = False
        for line in content:
            if line in NETSTAT_SECTION_ID:

                # this is a new section
                cur_section = NetstatSection(line)
                sections.append(cur_section)
                is_meta_data = True
                continue

            if cur_section:
                if is_meta_data:
                    cur_section.add_meta_data(line)
                    is_meta_data = False
                else:
                    cur_section.add_data(line)

        if not sections:
            raise ParseException("Found no section headers in content")

        self.data = dict((s.name, s._merge_data_index()) for s in sections)
        self.lines = dict((s.name, s.lines) for s in sections)
        self.datalist = dict((s.name, s.datalist) for s in sections)
        self._dataobjs = dict((s.name, s) for s in sections)

    @property
    def running_processes(self):
        """
        List all the running processes given in the Active Internet
        Connections part of the netstat output.

        Returns:
            set: set of process names (with spaces, as given in netstat output)
        """

        # Is it possible to have a machine that has no active connections?
        if ACTIVE_INTERNET_CONNECTIONS not in self.data:
            return set()
        return set(
            pg.split('/', 1)[1].strip()
            for pg in self.data[ACTIVE_INTERNET_CONNECTIONS]['PID/Program name']
            if '/' in pg
        )

    @property
    def listening_pid(self):
        """
        Find PIDs of all LISTEN processes

        Returns:
            dict: If any are found, they are returned in a dictionary
            following the format::

                {'pid': ("addr": ip_address, 'port': port, 'name': process_name)}
        """
        pids = {}
        # Is it possible to have a machine that has no active connections?
        if ACTIVE_INTERNET_CONNECTIONS not in self.datalist:
            return pids
        connlist = self.datalist[ACTIVE_INTERNET_CONNECTIONS]
        for line in connlist:
            if line['State'] != 'LISTEN':
                continue
            if not (':' in line['Local Address'] and '/' in line['PID/Program name']):
                continue
            addr, port = line['Local Address'].strip().split(":", 1)
            pid, name = line['PID/Program name'].strip().split('/', 1)
            pids[pid] = {'addr': addr, 'port': port, 'name': name}
        return pids

    def get_original_line(self, section_id, index):
        """
        Get the original netstat line that is stripped white spaces
        """
        if section_id not in self.data:
            return
        return self.lines[section_id][index]

    def search(self, **kwargs):
        """
        Search for rows in the data matching keywords in the search.

        This method searches both the active internet connections and
        active UNIX domain sockets.  If you only want to search one, specify
        the name via the ``search_list`` keyword, e.g.::

            from insights.parsers import Netstat, ACTIVE_UNIX_DOMAIN_SOCKETS
            conns.search(search_list=[ACTIVE_UNIX_DOMAIN_SOCKETS], State='LISTEN')

        The ``search_list`` can be either a list, or a string, containing
        one of the named constants defined in this module.  If
        ``search_list`` is not given, both the active internet connections
        and active UNIX domain sockets are searched, in that order.

        The results of the search are compiled into one list.  This allows
        you to search for all listening processes, whether for internet
        connections or UNIX sockets, by e.g.::

            conns.search(State__contains='LISTEN')

        This method uses the :py:func:`insights.parsers.keyword_search`
        function - see its documentation for a complete description of its
        keyword recognition capabilities.
        """
        if 'search_list' in kwargs:
            search_list = []
            if isinstance(kwargs['search_list'], list):
                # Compile a list from matching strings
                search_list = [l for l in kwargs['search_list'] if l in NETSTAT_SECTION_ID]
            elif (
                isinstance(kwargs['search_list'], str)
                and kwargs['search_list'] in NETSTAT_SECTION_ID
            ):
                # Just use this string
                search_list = [kwargs['search_list']]
            del kwargs['search_list']
        else:
            search_list = [ACTIVE_INTERNET_CONNECTIONS, ACTIVE_UNIX_DOMAIN_SOCKETS]
        if not search_list:
            # No valid search list?  No items.
            return []

        found = []
        for l in search_list:
            found.extend(keyword_search(self.datalist[l], parent=self._dataobjs[l], **kwargs))
        return found


@parser(Specs.netstat_i)
class Netstat_I(CommandParser):
    """
    Parse the ``netstat -i`` command output  to get interface traffic info
    such as "TX-OK" and "RX-OK".

    The output of `netstat -i` looks like::

        Kernel Interface table
        Iface       MTU Met    RX-OK RX-ERR RX-DRP RX-OVR    TX-OK TX-ERR TX-DRP TX-OVR Flg
        bond0      1500   0   845265      0      0      0     1753      0      0      0 BMmRU
        bond1      1500   0   842447      0      0      0     4233      0      0      0 BMmRU
        eth0       1500   0   422518      0      0      0     1703      0      0      0 BMsRU
        eth1       1500   0   422747      0      0      0       50      0      0      0 BMsRU
        eth2       1500   0   421192      0      0      0     3674      0      0      0 BMsRU
        eth3       1500   0   421255      0      0      0      559      0      0      0 BMsRU
        lo        65536   0        0      0      0      0        0      0      0      0 LRU

    Examples:
        >>> type(traf)
        <class 'insights.parsers.netstat.Netstat_I'>
        >>> traf.data[0]['Iface']  # A list of the interfaces and stats.
        'bond0'
        >>> 'bond0' in traf.group_by_iface  # A dictionary keyed on interface.
        True
        >>> 'enp0s25' in traf.group_by_iface
        False
        >>> 'MTU' in traf.group_by_iface['bond0']
        True
        >>> traf.group_by_iface['bond0']['MTU']  # as string
        '1500'
        >>> traf.group_by_iface['bond0']['RX-OK']
        '845265'
    """

    @property
    def group_by_iface(self):
        return self._group_by_iface

    def parse_content(self, content):
        self._group_by_iface = {}
        # heading_ignore is first line we _don't_ want to ignore...
        table = parse_delimited_table(content, heading_ignore=['Iface'])
        self.data = [dict((k, v) for (k, v) in item.items()) for item in table]
        for entry in self.data:
            self._group_by_iface[entry["Iface"]] = dict(
                (k, v) for (k, v) in entry.items() if k != 'Iface'
            )
        return


@parser(Specs.ss)
class SsTUPNA(CommandParser, list):
    """
    Parse the output of the ``ss -tupna`` command.

    Sample input data looks like::

        Netid  State   Recv-Q Send-Q Local Address:Port  Peer Address:Port   Process
        udp    UNCONN  0      0                  *:55898            *:*
        udp    UNCONN  0      0          127.0.0.1:904              *:*      users:(("rpc.statd",pid=29559,fd=7))
        udp    UNCONN  0      0                  *:111              *:*      users:(("rpcbind",pid=953,fd=9))
        udp    UNCONN  0      0                 :::37968           :::12345  users:(("rpc.statd",pid=29559,fd=10))
        tcp    LISTEN  0      128                *:111              *:*      users:(("rpcbind",pid=1139,fd=5),("systemd",pid=1,fd=41))

    Examples:

        >>> type(ss)
        <class 'insights.parsers.netstat.SsTUPNA'>
        >>> len(ss[1].keys())
        7
        >>> ss[0]['Local-Address-Port']
        '*:55898'
        >>> ss[0]['State']
        'UNCONN'
        >>> rpcbind = ss.get_service("rpcbind")  # All connections opened by rpcbind
        >>> len(rpcbind)
        2
        >>> rpcbind[0]['State']
        'UNCONN'
        >>> rpcbind[1]['State']
        'LISTEN'
        >>> rpcbind[0]['Process']
        'users:(("rpcbind",pid=953,fd=9))'
        >>> rpcbind[1]['Process']
        'users:(("rpcbind",pid=1139,fd=5),("systemd",pid=1,fd=41))'
        >>> using_55898 = ss.get_port("55898")  # Both local and peer port searched
        >>> len(using_55898)
        1
        >>> 'Process' in using_55898  # Not in dictionary if field not found
        False
        >>> rpcbind == ss.get_localport('111')  # Only local port or address searched
        True
    """

    SS_TABLE_HEADER = ["Netid State Recv-Q Send-Q Local-Address-Port Peer-Address-Port Process"]

    def parse_content(self, content):
        # Use headings without spaces and colons
        self.extend(parse_delimited_table(self.SS_TABLE_HEADER + content[1:]))
        del content

    def get_service(self, service):
        return [l for l in self if l.get("Process", None) and service in l["Process"]]

    def get_localport(self, port):
        list_conn = []
        for line in self:
            local_port = line.get('Local-Address-Port')
            if (
                local_port
                and ':*' not in local_port
                and int(port) == int(local_port.split(':')[-1])
            ):
                list_conn.append(line)
        return list_conn

    def get_peerport(self, port):
        list_conn = []
        for line in self:
            peer_port = line.get('Peer-Address-Port')
            if peer_port and ':*' not in peer_port and int(port) == int(peer_port.split(':')[-1]):
                list_conn.append(line)
        return list_conn

    def get_port(self, port):
        return self.get_localport(port) + self.get_peerport(port)

    @property
    def data(self):
        """
        .. warning::

            Attribute "data" is deprecated. The parser result is a list instead.
        """
        return self


@parser(Specs.ss)
class SsTULPN(SsTUPNA):
    """
    .. warning::
        This parser is deprecated, please use :py:class:`SsTUPNA` instead.

    Parse the output of the ``ss -tulpn`` command which keeps the lines contain
    "UNCONN" or "LISTEN" in the output of ``ss -tupna`` command.

    Refer to the super class :py:class:`Ss` for details.
    """

    def __init__(self, *args, **kwargs):
        deprecated(SsTULPN, "Please use the :py:class:`SsTUPNA` instead.", "3.8.0")
        super(SsTULPN, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        # Keep the lines of UNCONN and LISTEN only
        content = [line for line in content if (('UNCONN' in line) or ('LISTEN' in line))]
        self.extend(parse_delimited_table(self.SS_TABLE_HEADER + content))
        del content


@parser(Specs.proc_netstat)
class ProcNsat(Parser):
    """
    Parse the content of the ``/proc/net/netstat`` file

    Sample input data looks like::

        TcpExt: SyncookiesSent SyncookiesRecv SyncookiesFailed EmbryonicRsts PruneCalled RcvPruned OfoPruned OutOfWindowIcmps LockDroppedIcmps ArpFilter TW TWRecycled TWKilled PAWSPassive PAWSActive PAWSEstab DelayedACKs DelayedACKLocked DelayedACKLost ListenOverflows ListenDrops TCPPrequeued TCPDirectCopyFromBacklog TCPDirectCopyFromPrequeue TCPPrequeueDropped TCPHPHits TCPHPHitsToUser TCPPureAcks TCPHPAcks TCPRenoRecovery TCPSackRecovery TCPSACKReneging TCPFACKReorder TCPSACKReorder TCPRenoReorder TCPTSReorder TCPFullUndo TCPPartialUndo TCPDSACKUndo TCPLossUndo TCPLostRetransmit TCPRenoFailures TCPSackFailures TCPLossFailures TCPFastRetrans TCPForwardRetrans TCPSlowStartRetrans TCPTimeouts TCPLossProbes TCPLossProbeRecovery TCPRenoRecoveryFail TCPSackRecoveryFail TCPSchedulerFailed TCPRcvCollapsed TCPDSACKOldSent TCPDSACKOfoSent TCPDSACKRecv TCPDSACKOfoRecv TCPAbortOnData TCPAbortOnClose TCPAbortOnMemory TCPAbortOnTimeout TCPAbortOnLinger TCPAbortFailed TCPMemoryPressures TCPSACKDiscard TCPDSACKIgnoredOld TCPDSACKIgnoredNoUndo TCPSpuriousRTOs TCPMD5NotFound TCPMD5Unexpected TCPSackShifted TCPSackMerged TCPSackShiftFallback TCPBacklogDrop PFMemallocDrop TCPMinTTLDrop TCPDeferAcceptDrop IPReversePathFilter TCPTimeWaitOverflow TCPReqQFullDoCookies TCPReqQFullDrop TCPRetransFail TCPRcvCoalesce TCPOFOQueue TCPOFODrop TCPOFOMerge TCPChallengeACK TCPSYNChallenge TCPFastOpenActive TCPFastOpenActiveFail TCPFastOpenPassive TCPFastOpenPassiveFail TCPFastOpenListenOverflow TCPFastOpenCookieReqd TCPSpuriousRtxHostQueues BusyPollRxPackets TCPAutoCorking TCPFromZeroWindowAdv TCPToZeroWindowAdv TCPWantZeroWindowAdv TCPSynRetrans TCPOrigDataSent TCPHystartTrainDetect TCPHystartTrainCwnd TCPHystartDelayDetect TCPHystartDelayCwnd TCPACKSkippedSynRecv TCPACKSkippedPAWS TCPACKSkippedSeq TCPACKSkippedFinWait2 TCPACKSkippedTimeWait TCPACKSkippedChallenge TCPWqueueTooBig
        TcpExt: 10 20 30 40 0 0 0 0 0 0 8387793 2486 0 0 0 3 27599330 35876 309756 0 0 84351589 9652226708 54271044841 0 10507706759 112982361 177521295 3326559442 0 26212 0 36 33090 0 14345 959 8841 425 833 399 0 160 2 633809 11063 7056 233144 1060065 640242 0 228 54 0 310709 0 820887 112 900268 31664 0 232144 0 0 0 261 1048 808390 9 0 0 120433 244126 450077 0 0 0 5625 0 0 0 0 0 6772744900 19251701 0 0 465 463 0 0 0 0 0 0 1172 0 623074473 51282 51282 142025 465090 8484708872 836920 18212118 88 4344 0 0 5 4 3 2 1
        IpExt: InNoRoutes InTruncatedPkts InMcastPkts OutMcastPkts InBcastPkts OutBcastPkts InOctets OutOctets InMcastOctets OutMcastOctets InBcastOctets OutBcastOctets InCsumErrors InNoECTPkts InECT1Pkts InECT0Pkts InCEPkts ReasmOverlaps
        IpExt: 100 200 300 400 500 0 10468977960762 8092447661930 432 0 3062938 0 0 12512350267 400 300 200 100

    Examples:

        >>> type(pnstat)
        <class 'insights.parsers.netstat.ProcNsat'>
        >>> len(pnstat.data) == 132
        True
        >>> pnstat.get_stats('ReasmOverlaps')
        100
        >>> pnstat.get_stats('EmbryonicRsts')
        40
    """

    def parse_content(self, content):
        self.data = {}
        if not content:
            raise SkipComponent("No Contents")
        tcp_hdr = content[0].split()[1:]
        tcp_hdr_len = len(tcp_hdr)
        tcp_stat = content[1].split()[1:]
        tcp_stat_len = len(tcp_stat)
        ip_hdr = content[2].split()[1:]
        ip_hdr_len = len(ip_hdr)
        ip_stat = content[3].split()[1:]
        ip_stat_len = len(ip_stat)
        if (tcp_hdr_len != tcp_stat_len) or (ip_hdr_len != ip_stat_len):
            raise ParseException("Invalid Contents")
        if tcp_hdr and tcp_stat:
            self.data = dict(zip(tcp_hdr, tcp_stat))
        if ip_hdr and ip_stat:
            self.data.update(dict(zip(ip_hdr, ip_stat)))

    def get_stats(self, key_stats):
        """
        (int): The parser method will return the integer stats of the key if key is present in
               TcpExt or IpExt else it will return `None`.
        """
        if key_stats in self.data:
            return int(self.data.get(key_stats))



import doctest
import pytest

from insights.core.exceptions import ContentException, SkipComponent
from insights.parsers import postconf
from insights.parsers.postconf import PostconfBuiltin, Postconf, _Postconf
from insights.tests import context_wrap

V_OUT1 = """
""".strip()

V_OUT2 = """
smtpd_tls_loglevel = 0
smtpd_tls_mandatory_ciphers = medium
smtpd_tls_mandatory_exclude_ciphers =
smtpd_tls_mandatory_protocols = !SSLv2, !SSLv3, !TLSv1
""".strip()

V_OUT3 = """
command not found
""".strip()


def test_PostconfBuiltin():
    with pytest.raises(SkipComponent):
        PostconfBuiltin(context_wrap(V_OUT1))

    with pytest.raises(ContentException):
        PostconfBuiltin(context_wrap(V_OUT3))

    p = PostconfBuiltin(context_wrap(V_OUT2))
    assert p['smtpd_tls_loglevel'] == '0'
    assert p['smtpd_tls_mandatory_ciphers'] == 'medium'
    assert p['smtpd_tls_mandatory_exclude_ciphers'] == ''
    assert p['smtpd_tls_mandatory_protocols'] == '!SSLv2, !SSLv3, !TLSv1'


def test_Postconf():
    with pytest.raises(SkipComponent):
        Postconf(context_wrap(V_OUT1))

    with pytest.raises(ContentException):
        Postconf(context_wrap(V_OUT3))

    p = Postconf(context_wrap(V_OUT2))
    assert p['smtpd_tls_loglevel'] == '0'
    assert p['smtpd_tls_mandatory_ciphers'] == 'medium'
    assert p['smtpd_tls_mandatory_exclude_ciphers'] == ''
    assert p['smtpd_tls_mandatory_protocols'] == '!SSLv2, !SSLv3, !TLSv1'


def test_empty():
    with pytest.raises(SkipComponent):
        PostconfBuiltin(context_wrap(""))
    with pytest.raises(SkipComponent):
        Postconf(context_wrap(""))


def test_invalid():
    with pytest.raises(SkipComponent):
        PostconfBuiltin(context_wrap("asdf"))
    with pytest.raises(SkipComponent):
        Postconf(context_wrap("asdf"))


def test_doc_examples():
    env = {
        'postconfb': PostconfBuiltin(context_wrap(V_OUT2)),
        'postconf': Postconf(context_wrap(V_OUT2)),
        '_postconf': _Postconf(context_wrap(V_OUT2)),
    }
    failed, total = doctest.testmod(postconf, globs=env)
    assert failed == 0

    # TODO
    # env = {
    #     'postconf': Postconf(context_wrap(V_OUT2)),
    # }
    # failed, total = doctest.testmod(postconf, globs=env)
    # assert failed == 0



"""
Grubby - command ``/usr/sbin/grubby``
=====================================

This is a collection of parsers that all deal with the command ``grubby``.
Parsers included in this module are:

GrubbyDefaultIndex - command ``grubby --default-index``
-------------------------------------------------------

GrubbyDefaultKernel - command ``grubby --default-kernel``
---------------------------------------------------------

GrubbyInfoAll - command ``grubby --info=ALL``
---------------------------------------------
"""

from insights.core import CommandParser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsers import parse_cmdline_args
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.grubby_default_index)
class GrubbyDefaultIndex(CommandParser):
    """
    This parser parses the output of command ``grubby --default-index``.

    The typical output of this command is::

        0

    Examples:
        >>> grubby_default_index.default_index
        0

    Raises:
        SkipComponent: When output is empty
        ParseException: When output is invalid

    Attributes:
        default_index (int): the numeric index of the current default boot entry, count from 0
        error_lines (list): the error messages from the grubby command
    """

    def parse_content(self, content):
        if not content:
            raise SkipComponent('Empty output')

        self.error_lines = []
        default_index_line = content[0]
        if len(content) != 1:
            default_index_line = content[-1]
            self.error_lines = content[:-1]

        if not default_index_line.strip().isdigit():
            raise ParseException('Invalid output: {0}'.format(content))

        self.default_index = int(default_index_line.strip())


@parser(Specs.grubby_default_kernel)
class GrubbyDefaultKernel(CommandParser):
    """
    .. warning::
        This class is deprecated and will be removed from 3.7.0.
        Please use the :class:`insights.combiners.grubby.Grubby` instead.

    This parser parses the output of command ``grubby --default-kernel``.

    The typical output of this command is::

        /boot/vmlinuz-2.6.32-573.el6.x86_64

    Examples:

        >>> grubby_default_kernel.default_kernel
        '/boot/vmlinuz-2.6.32-573.el6.x86_64'

    Raises:
        SkipComponent: When output is empty
        ParseException: When output is invalid

    Attributes:
        default_kernel(str): The default kernel name for next boot
    """

    def __init__(self, context):
        deprecated(
            GrubbyDefaultKernel,
            "Please use the :class:`insights.combiners.grubby.Grubby` instead.",
            "3.7.0",
        )
        super(GrubbyDefaultKernel, self).__init__(context)

    def parse_content(self, content):
        if not content:
            raise SkipComponent('Empty output')

        # Skip the error lines to find the real default-kernel line.
        # Typically, the default-kernel line is the last line in content.
        default_kernel_str = None

        if len(content) == 1:
            default_kernel_str = content[0].strip()
        else:
            for idx in range(len(content) - 1, -1, -1):
                line = content[idx]
                if line.startswith('/boot/') and '/boot/vmlinuz-' in line:
                    if not default_kernel_str:
                        default_kernel_str = line.strip()
                    else:
                        raise ParseException(
                            'Invalid output: duplicate kernel lines: {0}'.format(line)
                        )

        if not default_kernel_str:
            raise ParseException('Invalid output: no kernel line: {0}'.format(content))
        if len(default_kernel_str.split()) > 1:
            raise ParseException(
                'Invalid output: unparsable kernel line: {0}'.format(default_kernel_str)
            )

        self.default_kernel = default_kernel_str


@parser(Specs.grubby_info_all)
class GrubbyInfoAll(CommandParser):
    """
    This parser parses the output of command ``grubby --info=ALL``.

    Attributes:
        boot_entries (dict): All boot entries indexed by the entry "index"
        unparsed_lines (list): All the unparsable lines

    The typical output of this command is::

        index=0
        kernel="/boot/vmlinuz-5.14.0-162.6.1.el9_1.x86_64"
        args="ro crashkernel=1G-4G:192M,4G-64G:256M,64G-:512M resume=/dev/mapper/rhel-swap rd.lvm.lv=rhel/root rd.lvm.lv=rhel/swap rhgb quiet retbleed=stuff"
        root="/dev/mapper/rhel-root"
        initrd="/boot/initramfs-5.14.0-162.6.1.el9_1.x86_64.img"
        title="Red Hat Enterprise Linux (5.14.0-162.6.1.el9_1.x86_64) 9.1 (Plow)"
        id="4d684a4a6166439a867e701ded4f7e10-5.14.0-162.6.1.el9_1.x86_64"
        index=1
        kernel="/boot/vmlinuz-5.14.0-70.13.1.el9_0.x86_64"
        args="ro crashkernel=1G-4G:192M,4G-64G:256M,64G-:512M resume=/dev/mapper/rhel-swap rd.lvm.lv=rhel/root rd.lvm.lv=rhel/swap rhgb quiet retbleed=stuff"
        root="/dev/mapper/rhel-root"
        initrd="/boot/initramfs-5.14.0-70.13.1.el9_0.x86_64.img"
        title="Red Hat Enterprise Linux (5.14.0-70.13.1.el9_0.x86_64) 9.0 (Plow)"
        id="4d684a4a6166439a867e701ded4f7e10-5.14.0-70.13.1.el9_0.x86_64"

    Examples:

        >>> len(grubby_info_all.boot_entries)
        2
        >>> grubby_info_all.boot_entries[0]["kernel"]
        '/boot/vmlinuz-5.14.0-162.6.1.el9_1.x86_64'
        >>> grubby_info_all.boot_entries[1].get("args").get("rd.lvm.lv")
        ['rhel/root', 'rhel/swap']

    Raises:
        SkipComponent: When output is empty
        ParseException: When output is invalid
    """

    def parse_content(self, content):

        if not content:
            raise SkipComponent("Empty output")

        self.boot_entries = {}
        self.unparsed_lines = []

        entry_data = {}
        for _line in content:
            line = _line.strip()

            if not line:
                continue
            if "=" not in line:
                self.unparsed_lines.append(_line)
                continue

            k, v = line.split("=", 1)
            v = v.strip("'\"")
            if k == "index":
                if v.isdigit():
                    if entry_data and "index" in entry_data and len(entry_data) > 1:
                        self.boot_entries[entry_data["index"]] = entry_data
                    entry_data = {k: int(v)}
                else:
                    raise ParseException('Invalid index value: {0}'.format(_line))
            elif k == "args":
                entry_data["raw_args"] = v
                entry_data[k] = parse_cmdline_args(v)
            else:
                entry_data[k] = v

        if entry_data and "index" in entry_data and len(entry_data) > 1:
            self.boot_entries[entry_data["index"]] = entry_data

        if not self.boot_entries:
            raise SkipComponent("No valid entry parsed")



import doctest

from insights.tests import context_wrap
from insights.parsers import satellite_yaml


SATELLITE_YAML_FILE_CONTENT = """
---
:answer_file: "/etc/foreman-installer/scenarios.d/satellite-answers.yaml"
:color_of_background: :dark
:colors: true
:custom:
  :lock_package_versions: true
:description: Install Satellite server
:dont_save_answers: false
:enabled: true
:facts:
  tuning: default
  mongo_cache_size: 3.89
:hiera_config: "/usr/share/foreman-installer/config/foreman-hiera.yaml"
:hook_dirs:
- "/usr/share/foreman-installer/katello/hooks"
:ignore_undocumented: false
:installer_dir: "/usr/share/foreman-installer"
:log_dir: "/var/log/foreman-installer"
:log_level: DEBUG
:log_name: satellite.log
:low_priority_modules: []
:mapping: {}
:module_dirs: "/usr/share/foreman-installer/modules"
:name: Satellite
:no_prefix: false
:order:
- certs
- foreman
- katello
- foreman_proxy
- foreman_proxy::plugin::pulp
- foreman_proxy_content
- puppet
:parser_cache_path:
- "/usr/share/foreman-installer/parser_cache/foreman.yaml"
- "/usr/share/foreman-installer/parser_cache/katello.yaml"
:skip_puppet_version_check: false
:store_dir: ''
:verbose: true
:verbose_log_level: notice
""".strip()

sat_yaml_path = "/etc/foreman-installer/scenarios.d/satellite.yaml"


def test_satellite_yaml():
    result = satellite_yaml.SatelliteYaml(
        context_wrap(SATELLITE_YAML_FILE_CONTENT, path=sat_yaml_path)
    )
    assert ':facts' in result
    assert result[':facts']['tuning'] == 'default'


def test_doc():
    failed_count, tests = doctest.testmod(
        satellite_yaml,
        globs={
            "SatelliteYaml": satellite_yaml.SatelliteYaml(context_wrap(SATELLITE_YAML_FILE_CONTENT))
        },
    )
    assert failed_count == 0
